import os
import sys
import posixpath
import time
import datetime
import glob
import shutil
import fnmatch
import zipfile
if os.name == 'nt':
    import msvcrt
elif os.name == 'posix':
    import fcntl
try:
    import cPickle as pickle
except ImportError:
    import pickle
try:
    import json as simplejson
except ImportError:
    import simplejson
import email
import email.Utils
import email.Generator
import email.Message
import email.header
import email.encoders
import smtplib
import ftplib
import socket
import ssl
from django.utils.translation import ugettext as _
#Bots modules
import botslib
import botsglobal
from botsconfig import *

@botslib.log_session
def run(idchannel,command,idroute,rootidta=None):
    '''run a communication session (dispatcher for communication functions).'''
    if rootidta is None:
        rootidta = botsglobal.currentrun.get_minta4query()
    for row in botslib.query('''SELECT *
                                FROM channel
                                WHERE idchannel=%(idchannel)s''',
                                {'idchannel':idchannel}):
        channeldict = dict(row)   #convert to real dictionary ()
        botsglobal.logger.debug(u'Start communication channel "%(idchannel)s" type %(type)s %(inorout)s.',channeldict)
        #for acceptance testing bots has an option to turn of external communication in channels
        if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
            #override values in channels for acceptance testing.
            #testpath is used to 'trigger' this: if testpath has value, use acceptance.
            if channeldict['testpath']:
                channeldict['path'] = channeldict['testpath']    #use the testpath to specify where to find acceptance  tests.
                channeldict['remove'] = False    #never remove during acceptance testing
                if channeldict['type'] in ['file','mimefile','trash']:
                    pass #do nothing, same type
                elif channeldict['type'] in ['smtp','smtps','smtpstarttls','pop3','pop3s','pop3apop','imap4','imap4s']:
                    channeldict['type'] = 'mimefile'
                else:   #channeldict['type'] in ['ftp','ftps','ftpis','sftp','xmlrpc','ftp','ftp','communicationscript','db',]
                    channeldict['type'] = 'file'
            botsglobal.logger.debug(u'Channel "%(idchannel)s" adapted for acceptance test: type "%(type)s", testpath "%(testpath)s".',channeldict)

        #update communication/run process with idchannel
        ta_run = botslib.OldTransaction(botslib._Transaction.processlist[-1])
        if channeldict['inorout'] == 'in':
            ta_run.update(fromchannel=channeldict['idchannel'])
        else:
            ta_run.update(tochannel=channeldict['idchannel'])

        try:
            userscript,scriptname = botslib.botsimport('communicationscripts',channeldict['idchannel'])
        except botslib.BotsImportError:       #communicationscript is not there; other errors like syntax errors are not catched
            userscript = scriptname = None
        #get the communication class to use:
        if userscript and hasattr(userscript,channeldict['type']):          #check communication class in userscript (sub classing)
            classtocall = getattr(userscript,channeldict['type'])
        elif userscript and hasattr(userscript,'UserCommunicationClass'):   #check for communication class called 'UserCommunicationClass' in userscript. 20110920: Obsolete, depreciated. Keep this for now.
            classtocall = getattr(userscript,'UserCommunicationClass')       #20130206: does have advantages...like for testing etc (no dependent upon type)
        else:
            classtocall = globals()[channeldict['type']]                    #get the communication class from this module

        comclass = classtocall(channeldict,idroute,userscript,scriptname,command,rootidta) #call the class for this type of channel
        comclass.run()
        botsglobal.logger.debug(u'Finished communication channel "%(idchannel)s" type %(type)s %(inorout)s.',channeldict)
        break   #there can only be one channel; this break takes care that if found, the 'else'-clause is skipped
    else:
        raise botslib.CommunicationError(_(u'Channel "%(idchannel)s" is unknown.'),{'idchannel':idchannel})


class _comsession(object):
    ''' Abstract class for communication-session. Use only subclasses.
        Subclasses are called by dispatcher function 'run'
        Often 'idroute' is passed as a parameter. This is ONLY because of the @botslib.log_session-wrapper!
        use self.idroute!!
    '''
    def __init__(self,channeldict,idroute,userscript,scriptname,command,rootidta):
        self.channeldict = channeldict
        self.idroute = idroute
        self.userscript = userscript
        self.scriptname = scriptname
        self.command = command
        self.rootidta = rootidta

    def run(self):
        if self.channeldict['inorout'] == 'out':
            self.precommunicate()
            self.connect()
            self.outcommunicate()
            self.disconnect()
            self.archive()
        else:   #incommunication
            if self.command == 'new': #only in-communicate for new run
                #handle maxsecondsperchannel: use global value from bots.ini unless specified in channel. (In database this is field 'rsrv2'.)
                self.maxsecondsperchannel = botsglobal.ini.getint('settings','maxsecondsperchannel',sys.maxint) if self.channeldict['rsrv2'] <= 0 else self.channeldict['rsrv2']
                try:
                    self.connect()
                except:     #in-connection failed. note that no files are received yet. useful if scheduled quite often, and you do nto want error-report eg when server is down.
                    #max_nr_retry : get this from channel. should be integer, but only textfields where left. so might be ''/None->use 0
                    max_nr_retry = int(self.channeldict['rsrv1']) if self.channeldict['rsrv1'] else 0
                    if max_nr_retry:
                        domain = u'bots_communication_failure_' + self.channeldict['idchannel']
                        nr_retry = botslib.unique(domain)  #update nr_retry in database
                        if nr_retry >= max_nr_retry:
                            botslib.unique(domain,updatewith=0)    #reset nr_retry to zero
                            #raises exception
                        else:
                            return  #just return if max_nr_retry is not reached
                    raise
                else:   #in-connection OK
                    #max_nr_retry : get this from channel. should be integer, but only textfields where left. so might be ''/None->use 0
                    max_nr_retry = int(self.channeldict['rsrv1']) if self.channeldict['rsrv1'] else 0
                    if max_nr_retry:
                        domain = u'bots_communication_failure_' + self.channeldict['idchannel']
                        botslib.unique(domain,updatewith=0)    #set nr_retry to zero
                self.incommunicate()
                self.disconnect()
            self.postcommunicate()
            self.archive()


    def archive(self):
        ''' after the communication channel has ran, archive received of send files.
            archivepath is the root directory for the archive (for this channel).
            within the archivepath files are stored by default as [archivepath]/[date]/[unique_filename]

        '''
        if not self.channeldict['archivepath']:
            return  #do not archive if not indicated
        if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
            return  #do not archive in acceptance testing
        if self.channeldict['filename'] and self.channeldict['type'] in ('file','ftp','ftps','ftpis','sftp','mimefile','communicationscript'):
            archiveexternalname = botsglobal.ini.getboolean('settings','archiveexternalname',False) #use external filename in archive
        else:
            archiveexternalname = False
        if self.channeldict['inorout'] == 'in':
            status = FILEIN
            statust = OK
            channel = 'fromchannel'
        else:
            if archiveexternalname:
                status = EXTERNOUT
            else:
                status = FILEOUT
            statust = DONE
            channel = 'tochannel'
        #user script can manipulate archivepath
        if self.userscript and hasattr(self.userscript,'archivepath'):
            archivepath = botslib.runscript(self.userscript,self.scriptname,'archivepath',channeldict=self.channeldict)
        else:
            archivepath = botslib.join(self.channeldict['archivepath'],time.strftime('%Y%m%d'))
        archivezip = botsglobal.ini.getboolean('settings','archivezip',False)   #archive to zip or not
        if archivezip:
            archivepath += '.zip'
        checkedifarchivepathisthere = False  #for a outchannel that is less used, lots of empty dirs will be created. This var is used to check within loop if dir exist, but this is only checked one time.
        for row in botslib.query('''SELECT filename,idta
                                    FROM  ta
                                    WHERE idta>%(rootidta)s
                                    AND   status=%(status)s
                                    AND   statust=%(statust)s
                                    AND   ''' + channel + '''=%(idchannel)s
                                    ''',
                                    {'idchannel':self.channeldict['idchannel'],'status':status,
                                    'statust':statust,'rootidta':self.rootidta}):
            if not checkedifarchivepathisthere:
                if archivezip:
                    botslib.dirshouldbethere(os.path.dirname(archivepath))
                    archivezipfilehandler = zipfile.ZipFile(archivepath,'a',zipfile.ZIP_DEFLATED)
                else:
                    botslib.dirshouldbethere(archivepath)
                checkedifarchivepathisthere = True

            if archiveexternalname:
                if self.channeldict['inorout'] == 'in':
                    # we have internal filename, get external
                    absfilename = botslib.abspathdata(row['filename'])
                    taparent = botslib.OldTransaction(idta=row['idta'])
                    ta_list = botslib.trace_origin(ta=taparent,where={'status':EXTERNIN})
                    if ta_list:
                        archivename = os.path.basename(ta_list[-1].filename)
                    else:
                        archivename = row['filename']
                else:
                    # we have external filename, get internal
                    archivename = os.path.basename(row['filename'])
                    taparent = botslib.OldTransaction(idta=row['idta'])
                    ta_list = botslib.trace_origin(ta=taparent,where={'status':FILEOUT})
                    absfilename = botslib.abspathdata(ta_list[0].filename)
            else:
                # use internal name in archive
                absfilename = botslib.abspathdata(row['filename'])
                archivename = os.path.basename(row['filename'])

            if self.userscript and hasattr(self.userscript,'archivename'):
                archivename = botslib.runscript(self.userscript,self.scriptname,'archivename',channeldict=self.channeldict,idta=row['idta'],filename=absfilename)
            #~print 'archive',os.path.basename(absfilename),'as',archivename

            if archivezip:
                archivezipfilehandler.write(absfilename,archivename)
            else:
                # if a file of the same name already exists, add a timestamp
                if os.path.isfile(botslib.join(archivepath,archivename)):
                    archivename = os.path.splitext(archivename)[0] + time.strftime('_%H%M%S') + os.path.splitext(archivename)[1]
                shutil.copy(absfilename,botslib.join(archivepath,archivename))

        if archivezip and checkedifarchivepathisthere:
            archivezipfilehandler.close()


    def postcommunicate(self):
        pass

    def precommunicate(self):
        pass

    def file2mime(self):
        ''' convert 'plain' files into email (mime-document).
            1 edi file always in 1 mail.
            from status FILEOUT to FILEOUT
        '''
        #select files with right statust, status and channel.
        for row in botslib.query('''SELECT idta,filename,frompartner,topartner,charset,contenttype,editype
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND tochannel=%(idchannel)s
                                    ''',
                                    {'idchannel':self.channeldict['idchannel'],'status':FILEOUT,
                                    'statust':OK,'idroute':self.idroute,'rootidta':self.rootidta}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=FILEOUT)
                ta_to.synall()  #needed for user exits: get all parameters of ta_to from database;
                confirmtype = u''
                confirmasked = False
                charset = row['charset']

                if row['editype'] == 'email-confirmation': #outgoing MDN: message is already assembled
                    outfilename = row['filename']
                else:   #assemble message: headers and payload. Bots uses simple MIME-envelope; by default payload is an attachment
                    message = email.Message.Message()
                    #set 'from' header (sender)
                    frommail,ccfrom_not_used_variable = self.idpartner2mailaddress(row['frompartner'])    #lookup email address for partnerID
                    message.add_header('From', frommail)

                    #set 'to' header (receiver)
                    if self.userscript and hasattr(self.userscript,'getmailaddressforreceiver'):    #user exit to determine to-address/receiver
                        tomail,ccto = botslib.runscript(self.userscript,self.scriptname,'getmailaddressforreceiver',channeldict=self.channeldict,ta=ta_to)
                    else:
                        tomail,ccto = self.idpartner2mailaddress(row['topartner'])          #lookup email address for partnerID
                    message.add_header('To',tomail)
                    if ccto:
                        message.add_header('CC',ccto)

                    if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
                        reference = '123message-ID email should be unique123'
                        email_datetime = email.Utils.formatdate(timeval=time.mktime(time.strptime("2013-01-23 01:23:45", "%Y-%m-%d %H:%M:%S")),localtime=True)
                    else:
                        reference = email.Utils.make_msgid(unicode(ta_to.idta))    #use transaction idta in message id.
                        email_datetime = email.Utils.formatdate(localtime=True)
                    message.add_header('Message-ID',reference)
                    message.add_header("Date",email_datetime)
                    ta_to.update(frommail=frommail,tomail=tomail,cc=ccto,reference=reference)   #update now (in order to use correct & updated ta_to in userscript)

                    #set Disposition-Notification-To: ask/ask not a a MDN?
                    if botslib.checkconfirmrules('ask-email-MDN',idroute=self.idroute,idchannel=self.channeldict['idchannel'],
                                                                frompartner=row['frompartner'],topartner=row['topartner']):
                        message.add_header("Disposition-Notification-To",frommail)
                        confirmtype = u'ask-email-MDN'
                        confirmasked = True

                    #set subject
                    if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
                        subject = u'12345678'
                    else:
                        subject = unicode(row['idta'])
                    content = botslib.readdata(row['filename'])     #get attachment from data file
                    if self.userscript and hasattr(self.userscript,'subject'):    #user exit to determine subject
                        subject = botslib.runscript(self.userscript,self.scriptname,'subject',channeldict=self.channeldict,ta=ta_to,subjectstring=subject,content=content)
                    message.add_header('Subject',subject)

                    #set MIME-version
                    message.add_header('MIME-Version','1.0')

                    #set attachment filename
                    filename_mask = self.channeldict['filename'] if self.channeldict['filename'] else '*'
                    attachmentfilename = self.filename_formatter(filename_mask,ta_to)
                    if attachmentfilename and self.channeldict['sendmdn'] != 'body':  #if not explicitly indicated 'as body' or (old)  if attachmentfilename is None or empty string: do not send as an attachment.
                        message.add_header("Content-Disposition",'attachment',filename=attachmentfilename)

                    #set Content-Type and charset
                    charset = self.convertcodecformime(row['charset'])
                    message.add_header('Content-Type',row['contenttype'].lower(),charset=charset)          #contenttype is set in grammar.syntax

                    #set attachment/payload; the Content-Transfer-Encoding is set by python encoder
                    message.set_payload(content)   #do not use charset; this lead to unwanted encodings...bots always uses base64
                    if self.channeldict['askmdn'] == 'never':       #channeldict['askmdn'] is the Mime encoding
                        email.encoders.encode_7or8bit(message)      #no encoding; but the Content-Transfer-Encoding is set to 7-bit or 8-bt
                    elif self.channeldict['askmdn'] == 'ascii' and charset == 'us-ascii':
                        pass        #do nothing: ascii is default encoding
                    else:           #if Mime encoding is 'always' or  (Mime encoding == 'ascii' and charset!='us-ascii'): use base64
                        email.encoders.encode_base64(message)

                    #*******write email to file***************************
                    outfilename = unicode(ta_to.idta)
                    outfile = botslib.opendata(outfilename, 'wb')
                    generator = email.Generator.Generator(outfile, mangle_from_=False, maxheaderlen=78)
                    generator.flatten(message,unixfrom=False)
                    outfile.close()
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt)
            else:
                ta_to.update(statust=OK,filename=outfilename,confirmtype=confirmtype,confirmasked=confirmasked,charset=charset)
            finally:
                ta_from.update(statust=DONE)
        return

    def mime2file(self):
        ''' convert emails (mime-documents) to 'plain' files.
            from status FILEIN to FILEIN
            process emails:
            -   extract information (eg sender-address)
            -   generate MDN (if asked and OK from bots-configuration)
            -   process MDN
            -   save 'attachments' as files
            -   filter emails/attachments based on contenttype
            -   email-address should be know by bots (can be turned off)
        '''
        whitelist_multipart = ['multipart/mixed','multipart/digest','multipart/signed','multipart/report','message/rfc822','multipart/alternative']
        whitelist_major = ['text','application']
        blacklist_contenttype = ['text/html','text/enriched','text/rtf','text/richtext','application/postscript','text/vcard','text/css']
        def savemime(msg):
            ''' save contents of email as separate files.
                is a nested function.
                3x filtering:
                -   whitelist of multipart-contenttype
                -   whitelist of body-contentmajor
                -   blacklist of body-contentytpe
            '''
            nrmimesaved = 0     #count nr of valid 'attachments'
            contenttype     = msg.get_content_type()
            if msg.is_multipart():
                if contenttype in whitelist_multipart:
                    for part in msg.get_payload():
                        nrmimesaved += savemime(part)
            else:    #is not a multipart
                if msg.get_content_maintype() not in whitelist_major or contenttype in blacklist_contenttype:
                    return 0
                content = msg.get_payload(decode=True)
                if not content or content.isspace():
                    return 0
                charset = msg.get_content_charset('ascii')
                if self.userscript and hasattr(self.userscript,'accept_incoming_attachment'):
                    accept_attachment = botslib.runscript(self.userscript,self.scriptname,'accept_incoming_attachment',channeldict=self.channeldict,ta=ta_from,charset=charset,content=content,contenttype=contenttype)
                    if not accept_attachment:
                        return 0
                filesize = len(content)
                ta_file = ta_from.copyta(status=FILEIN)
                outfilename = unicode(ta_file.idta)
                outfile = botslib.opendata(outfilename, 'wb')
                outfile.write(content)
                outfile.close()
                nrmimesaved += 1
                ta_file.update(statust=OK,
                                contenttype=contenttype,
                                charset=charset,
                                filename=outfilename,
                                filesize=filesize)
            return nrmimesaved
        #*****************end of nested function savemime***************************
        @botslib.log_session
        def mdnreceive():
            tmp = msg.get_param('reporttype')
            if tmp is None or email.Utils.collapse_rfc2231_value(tmp)!='disposition-notification':    #invalid MDN
                raise botslib.CommunicationInError(_(u'Received email-MDN with errors.'))
            for part in msg.get_payload():
                if part.get_content_type()=='message/disposition-notification':
                    originalmessageid = part['original-message-id']
                    if originalmessageid is not None:
                        break
            else:   #invalid MDN: 'message/disposition-notification' not in email
                raise botslib.CommunicationInError(_(u'Received email-MDN with errors.'))
            botslib.changeq('''UPDATE ta
                               SET confirmed=%(confirmed)s, confirmidta=%(confirmidta)s
                               WHERE reference=%(reference)s
                               AND status=%(status)s
                               AND confirmasked=%(confirmasked)s
                               AND confirmtype=%(confirmtype)s
                               ''',
                                {'status':FILEOUT,'reference':originalmessageid,'confirmed':True,'confirmtype':'ask-email-MDN','confirmidta':ta_from.idta,'confirmasked':True})
            #for now no checking if processing was OK.....
            #performance: not good. Index should be on the reference.
        @botslib.log_session
        def mdnsend(ta_from):
            if not botslib.checkconfirmrules('send-email-MDN',idroute=self.idroute,idchannel=self.channeldict['idchannel'],
                                                            frompartner=frompartner,topartner=topartner):
                return 0 #do not send
            #make message
            message = email.Message.Message()
            message.add_header('From',tomail)
            dispositionnotificationto = email.Utils.parseaddr(msg['disposition-notification-to'])[1]
            message.add_header('To', dispositionnotificationto)
            message.add_header('Subject', 'Return Receipt (displayed) - '+subject)
            message.add_header('MIME-Version','1.0')
            message.add_header('Content-Type','multipart/report',reporttype='disposition-notification')
            #~ message.set_type('multipart/report')
            #~ message.set_param('reporttype','disposition-notification')

            #make human readable message
            humanmessage = email.Message.Message()
            humanmessage.add_header('Content-Type', 'text/plain')
            humanmessage.set_payload('This is an return receipt for the mail that you send to '+tomail)
            message.attach(humanmessage)

            #make machine readable message
            machinemessage = email.Message.Message()
            machinemessage.add_header('Content-Type', 'message/disposition-notification')
            machinemessage.add_header('Original-Message-ID', reference)
            nep = email.Message.Message()
            machinemessage.attach(nep)
            message.attach(machinemessage)

            #write email to file;
            ta_mdn = botslib.NewTransaction(status=MERGED)  #new transaction for group-file

            if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
                mdn_reference = '123message-ID email should be unique123'
                mdn_datetime = email.Utils.formatdate(timeval=time.mktime(time.strptime("2013-01-23 01:23:45", "%Y-%m-%d %H:%M:%S")),localtime=True)
            else:
                mdn_reference = email.Utils.make_msgid(unicode(ta_mdn.idta))    #we first have to get the mda-ta to make this reference
                mdn_datetime = email.Utils.formatdate(localtime=True)
            message.add_header('Date',mdn_datetime)
            message.add_header('Message-ID', mdn_reference)

            mdnfilename = unicode(ta_mdn.idta)
            mdnfile = botslib.opendata(mdnfilename, 'wb')
            generator = email.Generator.Generator(mdnfile, mangle_from_=False, maxheaderlen=78)
            generator.flatten(message,unixfrom=False)
            mdnfile.close()
            ta_mdn.update(statust=OK,
                            idroute=self.idroute,
                            filename=mdnfilename,
                            editype='email-confirmation',
                            frompartner=topartner,
                            topartner=frompartner,
                            frommail=tomail,
                            tomail=dispositionnotificationto,
                            reference=mdn_reference,
                            content='multipart/report',
                            fromchannel=self.channeldict['idchannel'],
                            charset='ascii',
                            parent=ta_from.idta)
            return ta_mdn.idta
        #*****************end of nested function dispositionnotification***************************
        #get received mails for channel
        for row in botslib.query('''SELECT idta,filename
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND fromchannel=%(fromchannel)s
                                    ''',
                                    {'status':FILEIN,'statust':OK,'rootidta':self.rootidta,
                                    'fromchannel':self.channeldict['idchannel'],'idroute':self.idroute}):
            try:
                #default values for sending MDN; used to update ta if MDN is not asked
                confirmtype = ''
                confirmed = False
                confirmasked = False
                confirmidta = 0
                #read & parse email
                ta_from = botslib.OldTransaction(row['idta'])
                infile = botslib.opendata(row['filename'], 'rb')
                msg             = email.message_from_file(infile)   #read and parse mail
                infile.close()
                #******get information from email (sender, receiver etc)***********************************************************
                reference       = self.checkheaderforcharset(msg['message-id'])
                subject = self.checkheaderforcharset(msg['subject'])
                contenttype     = self.checkheaderforcharset(msg.get_content_type())
                #frompartner (incl autorization)
                frommail        = self.checkheaderforcharset(email.Utils.parseaddr(msg['from'])[1])
                frompartner = ''
                if not self.channeldict['starttls']:    #starttls in channeldict is: 'no check on "from:" email adress'
                    frompartner = self.mailaddress2idpartner(frommail)
                    if frompartner is None:
                        raise botslib.CommunicationInError(_(u'"From" emailaddress(es) %(email)s not authorised/unknown for channel "%(idchannel)s".'),
                                                            {'email':frommail,'idchannel':self.channeldict['idchannel']})
                #topartner, cc (incl autorization)
                list_to_address = [self.checkheaderforcharset(address) for name_not_used_variable,address in email.Utils.getaddresses(msg.get_all('to', []))]
                list_cc_address = [self.checkheaderforcharset(address) for name_not_used_variable,address in email.Utils.getaddresses(msg.get_all('cc', []))]
                cc_content      = ','.join([address for address in (list_to_address + list_cc_address)])
                topartner = ''  #initialise topartner
                tomail = ''     #initialise tomail
                if not self.channeldict['apop']:    #apop in channeldict is: 'no check on "to:" email adress'
                    for address in list_to_address:   #all tos-addresses are checked; only one needs to be authorised.
                        topartner =  self.mailaddress2idpartner(address)
                        tomail = address
                        if topartner is not None:   #if topartner found: break out of loop
                            break
                    else:   #if no valid topartner: generate error
                        raise botslib.CommunicationInError(_(u'"To" emailaddress(es) %(email)s not authorised/unknown for channel "%(idchannel)s".'),
                                                            {'email':list_to_address,'idchannel':self.channeldict['idchannel']})

                #update transaction of mail with information found in mail
                ta_from.update(frommail=frommail,   #why save now not later: when saving the attachments need the amil-header-info to be in ta (copyta)
                                tomail=tomail,
                                reference=reference,
                                contenttype=contenttype,
                                frompartner=frompartner,
                                topartner=topartner,
                                cc = cc_content,
                                rsrv1 = subject)
                if contenttype == 'multipart/report':   #process received MDN confirmation
                    mdnreceive()
                else:
                    if msg.has_key('disposition-notification-To'):  #sender requests a MDN
                        confirmidta = mdnsend(ta_from)
                        if confirmidta:
                            confirmtype = 'send-email-MDN'
                            confirmed = True
                            confirmasked = True
                    nrmimesaved = savemime(msg)
                    if not nrmimesaved:
                        raise botslib.CommunicationInError (_(u'No valid attachment in received email'))
            except:
                txt = botslib.txtexc()
                ta_from.update(statust=ERROR,errortext=txt)
                ta_from.deletechildren()
            else:
                ta_from.update(statust=DONE,confirmtype=confirmtype,confirmed=confirmed,confirmasked=confirmasked,confirmidta=confirmidta)
        return

    @staticmethod
    def checkheaderforcharset(org_header):
        ''' correct handling of charset for email headers that are saved in database.
        '''
        header,encoding = email.header.decode_header(org_header)[0]    #for subjects with non-ascii content special notation exists in MIME-standard
        try:
            if encoding is not None:
                return header.decode(encoding)                       #decode (to unicode)
            #test if valid; use case: spam... need to test because database-storage will give errors otherwise.
            header.encode('utf8')
            return header
        except:
            raise botslib.CommunicationInError(_(u'Email header invalid - probably issues with characterset.'))

    def mailaddress2idpartner(self,mailaddress):
        ''' lookup email address to see if know in configuration. '''
        mailaddress_lower = mailaddress.lower()
        #first check in chanpar email-addresses for this channel
        for row in botslib.query(u'''SELECT chanpar.idpartner_id as idpartner
                                    FROM chanpar,channel,partner
                                    WHERE chanpar.idchannel_id=channel.idchannel
                                    AND chanpar.idpartner_id=partner.idpartner
                                    AND partner.active=%(active)s
                                    AND chanpar.idchannel_id=%(idchannel)s
                                    AND LOWER(chanpar.mail)=%(mail)s''',
                                    {'active':True,'idchannel':self.channeldict['idchannel'],'mail':mailaddress_lower}):
            return row['idpartner']
        #if not found, check in partner-tabel (is less specific). Also test if in CC field.
        for row in botslib.query(u'''SELECT idpartner
                                    FROM partner
                                    WHERE active=%(active)s
                                    AND ( LOWER(mail) = %(mail)s OR LOWER(cc) LIKE %(maillike)s )''',
                                    {'active':True,'mail':mailaddress_lower,'maillike': '%' + mailaddress_lower + '%'}):
            return row['idpartner']
        return None     #indicate email address is unknown


    def idpartner2mailaddress(self,idpartner):
        for row in botslib.query(u'''SELECT chanpar.mail as mail,chanpar.cc as cc
                                    FROM chanpar,channel,partner
                                    WHERE chanpar.idchannel_id=channel.idchannel
                                    AND chanpar.idpartner_id=partner.idpartner
                                    AND partner.active=%(active)s
                                    AND chanpar.idchannel_id=%(idchannel)s
                                    AND chanpar.idpartner_id=%(idpartner)s''',
                                    {'active':True,'idchannel':self.channeldict['idchannel'],'idpartner':idpartner}):
            if row['mail']:
                return row['mail'],row['cc']
        for row in botslib.query(u'''SELECT mail,cc
                                    FROM partner
                                    WHERE active=%(active)s
                                    AND idpartner=%(idpartner)s''',
                                    {'active':True,'idpartner':idpartner}):
            if row['mail']:
                return row['mail'],row['cc']
        raise botslib.CommunicationOutError(_(u'No mail-address for partner "%(partner)s" (channel "%(idchannel)s").'),
                                                {'partner':idpartner,'idchannel':self.channeldict['idchannel']})

    def connect(self):
        pass

    def disconnect(self):
        pass

    @staticmethod
    def convertcodecformime(codec_in):
        convertdict = {
            'ascii' : 'us-ascii',
            'unoa' : 'us-ascii',
            'unob' : 'us-ascii',
            'unoc' : 'iso-8859-1',
            }
        codec_in = codec_in.lower().replace('_','-')
        return convertdict.get(codec_in,codec_in)


    def filename_formatter(self,filename_mask,ta):
        ''' Output filename generation from "template" filename configured in the channel
            Basically python's string.Formatter is used; see http://docs.python.org/library/string.html
            As in string.Formatter, substitution values are surrounded by braces; format specifiers can be used.
            Any ta value can be used
              eg. {botskey}, {alt}, {editype}, {messagetype}, {topartner}
            Next to the value in ta you can use:
            -   * : an unique number (per outchannel) using an asterisk
            -   {datetime}  use datetime with a valid strftime format:
                eg. {datetime:%Y%m%d}, {datetime:%H%M%S}
            -   {infile} use the original incoming filename; use name and extension, or either part separately:
                eg. {infile}, {infile:name}, {infile:ext}
            -   {overwrite}  if file wit hfielname exists: overwrite it (instead of appending)

            Exampels of usage:
                {botskey}_*.idoc        use incoming order number, add unique number, use extension '.idoc'
                *_{infile}              passthrough incoming filename & extension, prepend with unique number
                {infile:name}_*.txt     passthrough incoming filename, add unique number but change extension to .txt
                {editype}-{messagetype}-{datetime:%Y%m%d}-*.{infile:ext}
                                        use editype, messagetype, date and unique number with extension from the incoming file
                {topartner}/{messagetype}/*.edi
                                        Usage of subdirectories in the filename, they must already exist. In the example:
                                        sort into folders by partner and messagetype.

            Note1: {botskey} can only be used if merge is False for that messagetype
        '''
        class infilestr(str):
            ''' class for the infile-string that handles the specific format-options'''
            def __format__(self, format_spec):
                if not format_spec:
                    return unicode(self)
                name,ext = os.path.splitext(unicode(self))
                if format_spec == 'ext':
                    if ext.startswith('.'):
                        ext = ext[1:]
                    return ext
                if format_spec == 'name':
                    return name
                raise botslib.CommunicationOutError(_(u'Error in format of "{filename}": unknown format: "%(format)s".'),
                                                    {'format':format_spec})
        unique = unicode(botslib.unique(self.channeldict['idchannel'])) #create unique part for attachment-filename
        tofilename = filename_mask.replace('*',unique)           #filename_mask is filename in channel where '*' is replaced by idta
        if '{' in tofilename:    #only for python 2.6/7
            ta.synall()
            if '{infile' in tofilename:
                ta_list = botslib.trace_origin(ta=ta,where={'status':EXTERNIN})
                if ta_list:
                    infilename = infilestr(os.path.basename(ta_list[-1].filename))
                else:
                    infilename = ''
            else:
                infilename = ''
            try:
                if botsglobal.ini.getboolean('acceptance','runacceptancetest',False):
                    datetime_object = datetime.datetime.strptime("2013-01-23 01:23:45", "%Y-%m-%d %H:%M:%S")
                else:
                    datetime_object = datetime.datetime.now()
                tofilename = tofilename.format(infile=infilename,datetime=datetime_object,**ta.__dict__)
            except:
                txt = botslib.txtexc()
                raise botslib.CommunicationOutError(_(u'Error in formatting outgoing filename "%(filename)s". Error: "%(error)s".'),
                                                        {'filename':tofilename,'error':txt})
        if self.userscript and hasattr(self.userscript,'filename'):
            return botslib.runscript(self.userscript,self.scriptname,'filename',channeldict=self.channeldict,filename=tofilename,ta=ta)
        else:
            return tofilename


class file(_comsession):
    def connect(self):
        if self.channeldict['lockname']:        #directory locking: create lock-file. If the lockfile is already present an exception is raised.
            self.lockname = botslib.join(self.channeldict['path'],self.channeldict['lockname'])
            lock = os.open(self.lockname,os.O_WRONLY | os.O_CREAT | os.O_EXCL)
            os.close(lock)

    @botslib.log_session
    def incommunicate(self):
        ''' gets files from filesystem.
        '''
        frompath = botslib.join(self.channeldict['path'],self.channeldict['filename'])
        filelist = [filename for filename in glob.iglob(frompath) if os.path.isfile(filename)]
        filelist.sort()
        startdatetime = datetime.datetime.now()
        remove_ta = False
        for fromfilename in filelist:
            try:
                ta_from = botslib.NewTransaction(filename=fromfilename,
                                                status=EXTERNIN,
                                                fromchannel=self.channeldict['idchannel'],
                                                idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                #open fromfile, syslock if indicated
                fromfile = open(fromfilename,'rb')
                filesize = os.fstat(fromfile.fileno()).st_size
                if self.channeldict['syslock']:
                    if os.name == 'nt':
                        msvcrt.locking(fromfile.fileno(), msvcrt.LK_LOCK, 0x0fffffff)
                    elif os.name == 'posix':
                        fcntl.lockf(fromfile.fileno(), fcntl.LOCK_SH|fcntl.LOCK_NB)
                    else:
                        raise botslib.LockedFileError(_(u'Can not do a systemlock on this platform'))
                #open tofile
                tofilename = unicode(ta_to.idta)
                tofile = botslib.opendata(tofilename, 'wb')
                #copy
                shutil.copyfileobj(fromfile,tofile,1048576)
                tofile.close()
                fromfile.close()
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='file-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
                if self.channeldict['remove']:
                    os.remove(fromfilename)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

    @botslib.log_session
    def outcommunicate(self):
        ''' does output of files to filesystem. To be used via send-dispatcher.
            Output is either:
            1.  1 outputfile, messages are appended; filename is a fixed name
            2.  to directory; new file for each db-ta; if file exits: overwrite. File has to have a unique name.
        '''
        #check if output dir exists, else create it.
        outputdir = botslib.join(self.channeldict['path'])
        botslib.dirshouldbethere(outputdir)
        #get right filename_mask & determine if fixed name (append) or files with unique names
        filename_mask = self.channeldict['filename'] if self.channeldict['filename'] else '*'
        if '{overwrite}' in filename_mask:
            filename_mask = filename_mask.replace('{overwrite}','')
            mode = 'wb'
        else:
            mode = 'ab'
        #select the db-ta's for this channel
        for row in botslib.query(u'''SELECT idta,filename,numberofresends
                                       FROM ta
                                      WHERE idta>%(rootidta)s
                                        AND status=%(status)s
                                        AND statust=%(statust)s
                                        AND tochannel=%(tochannel)s
                                        ORDER BY idta
                                        ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:    #for each db-ta:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to =   ta_from.copyta(status=EXTERNOUT)
                #open tofile, incl syslock if indicated
                tofilename = self.filename_formatter(filename_mask,ta_from)
                tofilename = botslib.join(outputdir,tofilename)
                tofile = open(tofilename, mode)
                if self.channeldict['syslock']:
                    if os.name == 'nt':
                        msvcrt.locking(tofile.fileno(), msvcrt.LK_LOCK, 0x0fffffff)
                    elif os.name == 'posix':
                        fcntl.lockf(tofile.fileno(), fcntl.LOCK_EX|fcntl.LOCK_NB)
                    else:
                        raise botslib.LockedFileError(_(u'Can not do a systemlock on this platform'))
                #open fromfile
                fromfile = botslib.opendata(row['filename'], 'rb')
                #copy
                shutil.copyfileobj(fromfile,tofile,1048576)
                fromfile.close()
                tofile.close()
                #Rename filename after writing file.
                #Function: safe file writing: do not want another process to read the file while it is being written.
                #This is safe because file rename is atomic within same file system (?what about network shares?)
                if self.channeldict['mdnchannel']:
                    tofilename_old = tofilename
                    tofilename = botslib.rreplace(tofilename_old,self.channeldict['mdnchannel'])
                    os.rename(tofilename_old,tofilename)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename=tofilename,numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        #delete directory-lockfile
        if self.channeldict['lockname']:
            os.remove(self.lockname)


class pop3(_comsession):
    def connect(self):
        import poplib
        self.session = poplib.POP3(host=self.channeldict['host'],port=int(self.channeldict['port']))
        self.session.set_debuglevel(botsglobal.ini.getint('settings','pop3debug',0))    #if used, gives information about session (on screen), for debugging pop3
        self.session.user(self.channeldict['username'])
        self.session.pass_(self.channeldict['secret'])

    @botslib.log_session
    def incommunicate(self):
        ''' Fetch messages from Pop3-mailbox.
            A bad connection is tricky, because mails are actually only deleted on the server when QUIT is successful.
            A solution would be to connect, fetch, delete and quit for each mail, but this might introduce other problems.
            So: keep a list of idta received OK.
            If QUIT is not successful than delete these ta's
        '''
        self.listoftamarkedfordelete = []
        maillist = self.session.list()[1]     #get list of messages #alt: (response, messagelist, octets) = popsession.list()     #get list of messages
        startdatetime = datetime.datetime.now()
        remove_ta = False
        for mail in maillist:
            try:
                ta_from = botslib.NewTransaction(filename='pop3://'+self.channeldict['username']+'@'+self.channeldict['host'],
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                mailid = int(mail.split()[0])  #first 'word' is the message number/ID
                maillines = self.session.retr(mailid)[1]        #alt: (header, messagelines, octets) = popsession.retr(messageID)
                tofile = botslib.opendata(tofilename, 'wb')
                content = os.linesep.join(maillines)
                filesize = len(content)
                tofile.write(content)
                tofile.close()
                if self.channeldict['remove']:      #on server side mail is marked to be deleted. The pop3-server will actually delete the file if the QUIT commnd is receieved!
                    self.session.dele(mailid)
                    #add idta's of received mail in a list. If connection is not OK, QUIT command to POP3 server will not work. deleted mail will still be on server.
                    self.listoftamarkedfordelete += [ta_from.idta,ta_to.idta]

            except:         #something went wrong for this mail.
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='pop3-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
                #test connection. if connection is not OK stop fetching mails.
                try:
                    self.session.noop()
                except:
                    self.session = None     #indicate session is not valid anymore
                    break
            else:
                ta_to.update(statust=OK,filename=tofilename,filesize=filesize)
                ta_from.update(statust=DONE)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

    def disconnect(self):
        try:
            if not self.session:
                raise Exception(_(u'Pop3 connection not OK'))
            resp = self.session.quit()     #pop3 server will now actually delete the mails
            if resp[:1] != '+':
                raise Exception(_(u'QUIT command to POP3 server failed'))
        except Exception:   #connection is gone. Delete everything that is received to avoid double receiving.
            botslib.ErrorProcess(functionname='pop3-incommunicate',errortext='Could not fetch emails via POP3; probably communication problems',channeldict=self.channeldict)
            for idta in self.listoftamarkedfordelete:
                ta = botslib.OldTransaction(idta)
                ta.delete()

    @botslib.log_session
    def postcommunicate(self):
        self.mime2file()

class pop3s(pop3):
    def connect(self):
        import poplib
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        self.session = poplib.POP3_SSL(host=self.channeldict['host'],port=int(self.channeldict['port']),keyfile=keyfile,certfile=certfile)
        self.session.set_debuglevel(botsglobal.ini.getint('settings','pop3debug',0))    #if used, gives information about session (on screen), for debugging pop3
        self.session.user(self.channeldict['username'])
        self.session.pass_(self.channeldict['secret'])

class pop3apop(pop3):
    def connect(self):
        import poplib
        self.session = poplib.POP3(host=self.channeldict['host'],port=int(self.channeldict['port']))
        self.session.set_debuglevel(botsglobal.ini.getint('settings','pop3debug',0))    #if used, gives information about session (on screen), for debugging pop3
        self.session.apop(self.channeldict['username'],self.channeldict['secret'])    #python handles apop password encryption


class imap4(_comsession):
    ''' Fetch email from IMAP server.
    '''
    def connect(self):
        import imaplib
        imaplib.Debug = botsglobal.ini.getint('settings','imap4debug',0)    #if used, gives information about session (on screen), for debugging imap4
        self.session = imaplib.IMAP4(host=self.channeldict['host'],port=int(self.channeldict['port']))
        self.session.login(self.channeldict['username'],self.channeldict['secret'])

    @botslib.log_session
    def incommunicate(self):
        ''' Fetch messages from imap4-mailbox.
        '''
        # path may contain a mailbox name, otherwise use INBOX
        if self.channeldict['path']:
            mailbox_name = self.channeldict['path']
        else:
            mailbox_name = 'INBOX'

        response, data = self.session.select(mailbox_name)
        if response != 'OK': # eg. mailbox does not exist
            raise botslib.CommunicationError(mailbox_name + ': ' + data[0])

        # Get the message UIDs that should be read
        response, data = self.session.uid('search', None, '(UNDELETED)')
        if response != 'OK': # have never seen this happen, but just in case!
            raise botslib.CommunicationError(mailbox_name + ': ' + data[0])

        maillist = data[0].split()
        startdatetime = datetime.datetime.now()
        remove_ta = False
        for mail in maillist:
            try:
                ta_from = botslib.NewTransaction(filename='imap4://'+self.channeldict['username']+'@'+self.channeldict['host'],
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                filename = unicode(ta_to.idta)
                # Get the message (header and body)
                response, msg_data = self.session.uid('fetch',mail, '(RFC822)')
                filehandler = botslib.opendata(filename, 'wb')
                filesize = len(msg_data[0][1])
                filehandler.write(msg_data[0][1])
                filehandler.close()
                # Flag message for deletion AND expunge. Direct expunge has advantages for bad (internet)connections.
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='imap4-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            else:
                ta_to.update(statust=OK,filename=filename,filesize=filesize)
                ta_from.update(statust=DONE)
                if self.channeldict['remove']:
                    self.session.uid('store',mail, '+FLAGS', r'(\Deleted)')
                    self.session.expunge()
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

        self.session.close()        #Close currently selected mailbox. This is the recommended command before 'LOGOUT'.

    @botslib.log_session
    def postcommunicate(self):
        self.mime2file()

    def disconnect(self):
        self.session.logout()

class imap4s(imap4):
    def connect(self):
        import imaplib
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        imaplib.Debug = botsglobal.ini.getint('settings','imap4debug',0)    #if used, gives information about session (on screen), for debugging imap4
        self.session = imaplib.IMAP4_SSL(host=self.channeldict['host'],port=int(self.channeldict['port']),keyfile=keyfile,certfile=certfile)
        self.session.login(self.channeldict['username'],self.channeldict['secret'])


class smtp(_comsession):
    @botslib.log_session
    def precommunicate(self):
        self.file2mime()

    def connect(self):
        self.session = smtplib.SMTP(host=self.channeldict['host'],port=int(self.channeldict['port'])) #make connection
        self.session.set_debuglevel(botsglobal.ini.getint('settings','smtpdebug',0))    #if used, gives information about session (on screen), for debugging smtp
        self.login()

    def login(self):
        if self.channeldict['username'] and self.channeldict['secret']:
            try:
                #error in python 2.6.4....user and password can not be unicode
                self.session.login(str(self.channeldict['username']),str(self.channeldict['secret']))
            except smtplib.SMTPAuthenticationError:
                raise botslib.CommunicationOutError(_(u'SMTP server did not accept user/password combination.'))
            except:
                txt = botslib.txtexc()
                raise botslib.CommunicationOutError(_(u'SMTP login failed. Error:\n%(txt)s'),{'txt':txt})

    @botslib.log_session
    def outcommunicate(self):
        ''' does smtp-session.
            SMTP does not allow rollback. So if the sending of a mail fails, other mails may have been send.
        '''
        #send messages
        for row in botslib.query(u'''SELECT idta,filename,frommail,tomail,cc,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND tochannel=%(tochannel)s
                                    ''',
                                    {'status':FILEOUT,'statust':OK,'rootidta':self.rootidta,
                                    'tochannel':self.channeldict['idchannel']}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=EXTERNOUT)
                addresslist = row['tomail'].split(',') + row['cc'].split(',')
                addresslist = [x.strip() for x in addresslist if x.strip()]
                sendfile = botslib.opendata(row['filename'], 'rb')
                msg = sendfile.read()
                sendfile.close()
                self.session.sendmail(row['frommail'], addresslist, msg)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,filename='smtp://'+self.channeldict['username']+'@'+self.channeldict['host'],numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename='smtp://'+self.channeldict['username']+'@'+self.channeldict['host'],numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        try:    #Google gives/gave error closing connection. Not a real problem.
            self.session.quit()
        except socket.sslerror:     #for a starttls connection
            self.session.close()
        except:
            pass

class smtps(smtp):
    def connect(self):
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        self.session = smtplib.SMTP_SSL(host=self.channeldict['host'],port=int(self.channeldict['port']),keyfile=keyfile,certfile=certfile) #make connection
        self.session.set_debuglevel(botsglobal.ini.getint('settings','smtpdebug',0))    #if used, gives information about session (on screen), for debugging smtp
        self.login()

class smtpstarttls(smtp):
    def connect(self):
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        self.session = smtplib.SMTP(host=self.channeldict['host'],port=int(self.channeldict['port'])) #make connection
        self.session.set_debuglevel(botsglobal.ini.getint('settings','smtpdebug',0))    #if used, gives information about session (on screen), for debugging smtp
        self.session.ehlo()
        self.session.starttls(keyfile=keyfile,certfile=certfile)
        self.session.ehlo()
        self.login()


class mimefile(file):
    @botslib.log_session
    def postcommunicate(self):
        self.mime2file()
    @botslib.log_session
    def precommunicate(self):
        self.file2mime()

class ftp(_comsession):
    def connect(self):
        botslib.settimeout(botsglobal.ini.getint('settings','ftptimeout',10))
        self.session = ftplib.FTP()
        self.session.set_debuglevel(botsglobal.ini.getint('settings','ftpdebug',0))   #set debug level (0=no, 1=medium, 2=full debug)
        self.session.set_pasv(not self.channeldict['ftpactive']) #active or passive ftp
        self.session.connect(host=self.channeldict['host'],port=int(self.channeldict['port']))
        self.session.login(user=self.channeldict['username'],passwd=self.channeldict['secret'],acct=self.channeldict['ftpaccount'])
        self.set_cwd()

    def set_cwd(self):
        self.dirpath = self.session.pwd()
        if self.channeldict['path']:
            self.dirpath = posixpath.normpath(posixpath.join(self.dirpath,self.channeldict['path']))
            try:
                self.session.cwd(self.dirpath)           #set right path on ftp-server
            except:
                self.session.mkd(self.dirpath)           #set right path on ftp-server; no nested directories
                self.session.cwd(self.dirpath)           #set right path on ftp-server

    @botslib.log_session
    def incommunicate(self):
        ''' do ftp: receive files. To be used via receive-dispatcher.
            each to be imported file is transaction.
            each imported file is transaction.
        '''
        startdatetime = datetime.datetime.now()
        files = []
        try:            #some ftp servers give errors when directory is empty; catch these errors here
            files = self.session.nlst()
        except (ftplib.error_perm,ftplib.error_temp) as msg:
            if unicode(msg)[:3] not in [u'550',u'450']:
                raise

        lijst = fnmatch.filter(files,self.channeldict['filename'])
        remove_ta = False
        for fromfilename in lijst:  #fetch messages from ftp-server.
            try:
                ta_from = botslib.NewTransaction(filename='ftp:/'+posixpath.join(self.dirpath,fromfilename),
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],
                                                    idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                tofile = botslib.opendata(tofilename, 'wb')
                try:
                    if self.channeldict['ftpbinary']:
                        self.session.retrbinary("RETR " + fromfilename, tofile.write)
                    else:
                        self.session.retrlines("RETR " + fromfilename, lambda s, w=tofile.write: w(s+"\n"))
                except ftplib.error_perm as msg:
                    if unicode(msg)[:3] in [u'550',]:     #we are trying to download a directory...
                        raise botslib.BotsError(u'To be catched')
                    else:
                        raise
                tofile.close()
                filesize = os.path.getsize(botslib.abspathdata(tofilename))
                if not filesize:
                    raise botslib.BotsError(u'To be catched; directory (or empty file)')
            except botslib.BotsError:   #directory or empty file; handle exception but generate no error.
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='ftp-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
                if self.channeldict['remove']:
                    self.session.delete(fromfilename)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

    @botslib.log_session
    def outcommunicate(self):
        ''' do ftp: send files. To be used via receive-dispatcher.
            each to be send file is transaction.
            each send file is transaction.
            NB: ftp command APPE should be supported by server
        '''
        #get right filename_mask & determine if fixed name (append) or files with unique names
        filename_mask = self.channeldict['filename'] if self.channeldict['filename'] else '*'
        if '{overwrite}' in filename_mask:
            filename_mask = filename_mask.replace('{overwrite}','')
            mode = 'STOR '
        else:
            mode = 'APPE '
        for row in botslib.query('''SELECT idta,filename,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                      AND status=%(status)s
                                      AND statust=%(statust)s
                                      AND tochannel=%(tochannel)s
                                        ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=EXTERNOUT)
                tofilename = self.filename_formatter(filename_mask,ta_from)
                if self.channeldict['ftpbinary']:
                    fromfile = botslib.opendata(row['filename'], 'rb')
                    self.session.storbinary(mode + tofilename, fromfile)
                else:
                    fromfile = botslib.opendata(row['filename'], 'r')
                    self.session.storlines(mode + tofilename, fromfile)
                fromfile.close()
                #Rename filename after writing file.
                #Function: safe file writing: do not want another process to read the file while it is being written.
                if self.channeldict['mdnchannel']:
                    tofilename_old = tofilename
                    tofilename = botslib.rreplace(tofilename_old,self.channeldict['mdnchannel'])
                    self.session.rename(tofilename_old,tofilename)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,filename='ftp:/'+posixpath.join(self.dirpath,tofilename),numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename='ftp:/'+posixpath.join(self.dirpath,tofilename),numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        try:
            self.session.quit()
        except:
            self.session.close()
        botslib.settimeout(botsglobal.ini.getint('settings','globaltimeout',10))

class ftps(ftp):
    ''' explicit ftps as defined in RFC 2228 and RFC 4217.
        standard port to connect to is as in normal FTP (port 21)
        ftps is supported by python >= 2.7
    '''
    def connect(self):
        if not hasattr(ftplib,'FTP_TLS'):
            raise botslib.CommunicationError(_(u'ftps is not supported by your python version, use >=2.7'))
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        botslib.settimeout(botsglobal.ini.getint('settings','ftptimeout',10))
        self.session = ftplib.FTP_TLS(keyfile=keyfile,certfile=certfile)
        self.session.set_debuglevel(botsglobal.ini.getint('settings','ftpdebug',0))   #set debug level (0=no, 1=medium, 2=full debug)
        self.session.set_pasv(not self.channeldict['ftpactive']) #active or passive ftp
        self.session.connect(host=self.channeldict['host'],port=int(self.channeldict['port']))
        self.session.auth()
        self.session.login(user=self.channeldict['username'],passwd=self.channeldict['secret'],acct=self.channeldict['ftpaccount'])
        self.session.prot_p()
        self.set_cwd()


#sub classing of ftplib for ftpis
if hasattr(ftplib,'FTP_TLS'):
    class Ftp_tls_implicit(ftplib.FTP_TLS):
        ''' FTPS implicit is not directly supported by python; python>=2.7 supports only ftps explicit.
            So class ftplib.FTP_TLS is sub-classed here, with the needed modifications.
            (code is nicked from ftplib.ftp v. 2.7; additions/changes are indicated)
            '''
        def connect(self, host='', port=0, timeout=-999):
            #added hje 20110713: directly use SSL in FTPIS
            #end added
            if host != '':
                self.host = host
            if port > 0:
                self.port = port
            if timeout != -999:
                self.timeout = timeout
            self.sock = socket.create_connection((self.host, self.port), self.timeout)
            self.af = self.sock.family
            #added hje 20110713: directly use SSL in FTPIS
            self.sock = ssl.wrap_socket(self.sock, self.keyfile, self.certfile,ssl_version=self.ssl_version)
            #end added
            self.file = self.sock.makefile('rb')
            self.welcome = self.getresp()
            return self.welcome
        def prot_p(self):
            #Inovis FTPIS gives errors on 'PBSZ 0' and 'PROT P', vsftp does not work without these commands.
            #These errors are just catched, nothing is done with them.
            try:
                self.voidcmd('PBSZ 0')
            except ftplib.error_perm:
                pass
            try:
                resp = self.voidcmd('PROT P')
            except ftplib.error_perm:
                resp = None
            self._prot_p = True
            return resp


class ftpis(ftp):
    ''' FTPS implicit; is not defined in a RFC.
        standard port to connect is port 990.
        FTPS implicit is not supported by python.
        python>=2.7 supports ftps explicit.
        So used is the sub-class Ftp_tls_implicit.
        Tested with Inovis and VSFTPd.
        Python library FTP_TLS uses ssl_version = ssl.PROTOCOL_TLSv1
        Inovis seems to need PROTOCOL_SSLv3
        This is 'solved' by using 'parameters' in the channel.
        ~ ssl.PROTOCOL_SSLv2  = 0
        ~ ssl.PROTOCOL_SSLv3  = 1
        ~ ssl.PROTOCOL_SSLv23 = 2
        ~ ssl.PROTOCOL_TLSv1  = 3
    '''
    def connect(self):
        if not hasattr(ftplib,'FTP_TLS'):
            raise botslib.CommunicationError(_(u'ftpis is not supported by your python version, use >=2.7'))
        if self.userscript and hasattr(self.userscript,'keyfile'):
            keyfile, certfile = botslib.runscript(self.userscript,self.scriptname,'keyfile',channeldict=self.channeldict)
        elif self.channeldict['keyfile']:
            keyfile = self.channeldict['keyfile']
            certfile = self.channeldict['certfile']
        else:
            keyfile = certfile = None
        botslib.settimeout(botsglobal.ini.getint('settings','ftptimeout',10))
        self.session = Ftp_tls_implicit(keyfile=keyfile,certfile=certfile)
        if self.channeldict['parameters']:
            self.session.ssl_version = int(self.channeldict['parameters'])
        self.session.set_debuglevel(botsglobal.ini.getint('settings','ftpdebug',0))   #set debug level (0=no, 1=medium, 2=full debug)
        self.session.set_pasv(not self.channeldict['ftpactive']) #active or passive ftp
        self.session.connect(host=self.channeldict['host'],port=int(self.channeldict['port']))
        #~ self.session.auth()
        self.session.login(user=self.channeldict['username'],passwd=self.channeldict['secret'],acct=self.channeldict['ftpaccount'])
        self.session.prot_p()
        self.set_cwd()


class sftp(_comsession):
    ''' SFTP: SSH File Transfer Protocol (SFTP is not FTP run over SSH, SFTP is not Simple File Transfer Protocol)
        standard port to connect to is port 22.
        requires paramiko and pycrypto to be installed
        based on class ftp and ftps above with code from demo_sftp.py which is included with paramiko
        Mike Griffin 16/10/2010
        Henk-jan ebbers 20110802: when testing I found that the transport also needs to be closed. So changed transport ->self.transport, and close this in disconnect
        henk-jan ebbers 20111019: disabled the host_key part for now (but is very interesting). Is not tested; keys should be integrated in bots also for other protocols.
        henk-jan ebbers 20120522: hostkey and privatekey can now be handled in user exit.
    '''
    def connect(self):
        # check dependencies
        try:
            import paramiko
        except:
            raise ImportError(_(u'Dependency failure: communicationtype "sftp" requires python library "paramiko".'))
        # try:
            # from Crypto import Cipher
        # except:
            # raise ImportError(_(u'Dependency failure: communicationtype "sftp" requires python library "pycrypto".'))
        # setup logging if required
        ftpdebug = botsglobal.ini.getint('settings','ftpdebug',0)
        if ftpdebug > 0:
            log_file = botslib.join(botsglobal.ini.get('directories','logging'),'sftp.log')
            # Convert ftpdebug to paramiko logging level (1=20=info, 2=10=debug)
            paramiko.util.log_to_file(log_file, 30-(ftpdebug*10))

        # Get hostname and port to use
        hostname = self.channeldict['host']
        try:
            port = int(self.channeldict['port'])
        except:
            port = 22 # default port for sftp

        if self.userscript and hasattr(self.userscript,'hostkey'):
            hostkey = botslib.runscript(self.userscript,self.scriptname,'hostkey',channeldict=self.channeldict)
        else:
            hostkey = None
        if self.userscript and hasattr(self.userscript,'privatekey'):
            privatekeyfile,pkeytype,pkeypassword = botslib.runscript(self.userscript,self.scriptname,'privatekey',channeldict=self.channeldict)
            if pkeytype == 'RSA':
                pkey = paramiko.RSAKey.from_private_key_file(filename=privatekeyfile,password=pkeypassword)
            else:
                pkey = paramiko.DSSKey.from_private_key_file(filename=privatekeyfile,password=pkeypassword)
        else:
            pkey = None

        if self.channeldict['secret']:  #if password is empty string: use None. Else error can occur.
            secret = self.channeldict['secret']
        else:
            secret = None
        # now, connect and use paramiko Transport to negotiate SSH2 across the connection
        self.transport = paramiko.Transport((hostname,port))
        self.transport.connect(username=self.channeldict['username'],password=secret,hostkey=hostkey,pkey=pkey)
        self.session = paramiko.SFTPClient.from_transport(self.transport)
        channel = self.session.get_channel()
        channel.settimeout(botsglobal.ini.getint('settings','ftptimeout',10))
        self.set_cwd()

    def set_cwd(self):
        self.session.chdir('.') # getcwd does not work without this chdir first!
        self.dirpath = self.session.getcwd()
        if self.channeldict['path']:
            self.dirpath = posixpath.normpath(posixpath.join(self.dirpath,self.channeldict['path']))
            try:
                self.session.chdir(self.dirpath)
            except:
                self.session.mkdir(self.dirpath)
                self.session.chdir(self.dirpath)

    def disconnect(self):
        self.session.close()
        self.transport.close()


    @botslib.log_session
    def incommunicate(self):
        ''' do ftp: receive files. To be used via receive-dispatcher.
            each to be imported file is transaction.
            each imported file is transaction.
        '''
        startdatetime = datetime.datetime.now()
        files = self.session.listdir('.')
        lijst = fnmatch.filter(files,self.channeldict['filename'])
        remove_ta = False
        for fromfilename in lijst:  #fetch messages from sftp-server.
            try:
                ta_from = botslib.NewTransaction(filename='sftp:/'+posixpath.join(self.dirpath,fromfilename),
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],
                                                    idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                fromfile = self.session.open(fromfilename, 'r')    # SSH treats all files as binary
                content = fromfile.read()
                filesize = len(content)
                tofile = botslib.opendata(tofilename, 'wb')
                tofile.write(content)
                tofile.close()
                fromfile.close()
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='sftp-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
                if self.channeldict['remove']:
                    self.session.remove(fromfilename)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

    @botslib.log_session
    def outcommunicate(self):
        ''' do ftp: send files. To be used via receive-dispatcher.
            each to be send file is transaction.
            each send file is transaction.
        '''
        #get right filename_mask & determine if fixed name (append) or files with unique names
        filename_mask = self.channeldict['filename'] if self.channeldict['filename'] else '*'
        if '{overwrite}' in filename_mask:
            filename_mask = filename_mask.replace('{overwrite}','')
            mode = 'w'
        else:
            mode = 'a'
        for row in botslib.query('''SELECT idta,filename,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                      AND status=%(status)s
                                      AND statust=%(statust)s
                                      AND tochannel=%(tochannel)s
                                        ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=EXTERNOUT)
                tofilename = self.filename_formatter(filename_mask,ta_from)
                fromfile = botslib.opendata(row['filename'], 'rb')
                tofile = self.session.open(tofilename, mode)    # SSH treats all files as binary
                tofile.write(fromfile.read())
                tofile.close()
                fromfile.close()
                #Rename filename after writing file.
                #Function: safe file writing: do not want another process to read the file while it is being written.
                if self.channeldict['mdnchannel']:
                    tofilename_old = tofilename
                    tofilename = botslib.rreplace(tofilename_old,self.channeldict['mdnchannel'])
                    self.session.rename(tofilename_old,tofilename)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,filename='sftp:/'+posixpath.join(self.dirpath,tofilename),numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename='sftp:/'+posixpath.join(self.dirpath,tofilename),numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)


class xmlrpc(_comsession):
    ''' General xmlrpc implementation. Xmlrpc is often quite specific.
        Probably you will have to script your own xmlrpc class, but this is a good starting point.
        From channel is used: usernaem, secret, host, port, path. Path is the function to be used/called.
    '''
    def connect(self):
        import xmlrpclib
        uri = "http://%(username)s%(secret)s@%(host)s:%(port)s"%self.channeldict
        self.filename = "http://%(username)s@%(host)s:%(port)s"%self.channeldict    #used as 'filename' in reports etc
        session = xmlrpclib.ServerProxy(uri)
        self.xmlrpc_call = getattr(session,self.channeldict['path'])                #self.xmlrpc_call is called in communication

    @botslib.log_session
    def incommunicate(self):
        startdatetime = datetime.datetime.now()
        remove_ta = False
        while True:
            try:
                content = self.xmlrpc_call()
                if content is None:
                    break   #nothing (more) to receive.
                ta_from = botslib.NewTransaction(filename=self.filename,
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],
                                                    idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                tofile = botslib.opendata(tofilename, 'wb')
                simplejson.dump(content, tofile, skipkeys=False, ensure_ascii=False, check_circular=False)
                tofile.close()
                filesize = os.path.getsize(botslib.abspathdata(tofilename))
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='xmlprc-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
                break   #break out of while loop (else this would be endless)
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break

    @botslib.log_session
    def outcommunicate(self):
        ''' do xml-rpc: send files. To be used via receive-dispatcher.
            each to be send file is transaction.
            each send file is transaction.
        '''
        for row in botslib.query('''SELECT idta,filename,charset,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND tochannel=%(tochannel)s ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to =   ta_from.copyta(status=EXTERNOUT)
                fromfile = botslib.opendata(row['filename'], 'rb',row['charset'])
                content = fromfile.read()
                fromfile.close()
                response = self.xmlrpc_call(content)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename=self.filename,numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)


class db(_comsession):
    ''' communicate with a database; directly read or write from a database.
        the user HAS to provide a userscript file in usersys/communicationscripts that does the actual import/export using **some** python database library.
        the userscript file should contain:
        - connect
        - (for incoming) incommunicate
        - (for outgoing) outcommunicate
        - disconnect
        Other parameters are passed, use them for your own convenience.
        Bots 'pickles' the results returned from the userscript (and unpickles for the translation).
    '''
    def connect(self):
        if self.userscript is None:
            raise botslib.BotsImportError(_(u'Channel "%(idchannel)s" is type "db", but no communicationscript exists.'),
                                {'idchannel':self.channeldict['idchannel']})
        #check functions bots assumes to be present in userscript:
        if not hasattr(self.userscript,'connect'):
            raise botslib.ScriptImportError(_(u'No function "connect" in imported communicationscript "%(communicationscript)s".'),
                                                {'communicationscript':self.scriptname})
        if self.channeldict['inorout'] == 'in' and not hasattr(self.userscript,'incommunicate'):
            raise botslib.ScriptImportError(_(u'No function "incommunicate" in imported communicationscript "%(communicationscript)s".'),
                                                {'communicationscript':self.scriptname})
        if self.channeldict['inorout'] == 'out' and not hasattr(self.userscript,'outcommunicate'):
            raise botslib.ScriptImportError(_(u'No function "outcommunicate" in imported communicationscript "%(communicationscript)s".'),
                                                {'communicationscript':self.scriptname})
        if not hasattr(self.userscript,'disconnect'):
            raise botslib.ScriptImportError(_(u'No function "disconnect" in imported communicationscript "%(communicationscript)s".'),
                                            {'communicationscript':self.scriptname})

        self.dbconnection = botslib.runscript(self.userscript,self.scriptname,'connect',channeldict=self.channeldict)

    @botslib.log_session
    def incommunicate(self):
        ''' read data from database.
            userscript should return a 'db_objects'.
            This can be one edi-message or several edi-messages.
            if a list or tuple is passed: each element of list/tuple is treated as seperate edi-message.
            if this is None, do nothing
            if this is a list/tuple, each member of the list is send as a separate 'message'
            if you want all information from userscript to be passed as one edi message: pass as dict, eg {'data': <list of queries>}
        '''
        db_objects = botslib.runscript(self.userscript,self.scriptname,'incommunicate',channeldict=self.channeldict,dbconnection=self.dbconnection)
        if not db_objects:      #there should be a useful db_objects; if not just return (do nothing)
            return
        if not isinstance(db_objects,(list,tuple)):
            db_objects = [db_objects]   #a list or tuple is expected: pack received object in a list (list with one member).

        remove_ta = False
        for db_object in db_objects:
            try:
                ta_from = botslib.NewTransaction(filename=self.channeldict['path'],
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],
                                                    idroute=self.idroute)
                ta_to = ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                tofile = botslib.opendata(tofilename,'wb')
                pickle.dump(db_object, tofile)
                tofile.close()
                filesize = os.path.getsize(botslib.abspathdata(tofilename))
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='db-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
            finally:
                remove_ta = False

    @botslib.log_session
    def outcommunicate(self):
        ''' write data to database.
        '''
        for row in botslib.query('''SELECT idta,filename,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND tochannel=%(tochannel)s ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,'status':FILEOUT,'statust':OK}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=EXTERNOUT)
                fromfile = botslib.opendata(row['filename'], 'rb')
                db_object = pickle.load(fromfile)
                fromfile.close()
                botslib.runscript(self.userscript,self.scriptname,'outcommunicate',channeldict=self.channeldict,dbconnection=self.dbconnection,db_object=db_object)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,filename=self.channeldict['path'],numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename=self.channeldict['path'],numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        botslib.runscript(self.userscript,self.scriptname,'disconnect',channeldict=self.channeldict,dbconnection=self.dbconnection)



class communicationscript(_comsession):
    '''
    For running an userscript for communication.
    Examples of use:
    - call external communication program
    - call external program that extract messages from ERP-database
    - call external program that imports messages in ERP system
    - communication method not available in Bots ***or use sub-classing for this***
    - specialised I/O wishes; eg specific naming of output files. (eg including partner name) ***beter: use sub-classing or have more user exits***
    place of communicationscript: bots/usersys/communicationscripts
    name of communicationscript: same name as channel (the channelID)
    in this communicationscript some functions will be called:
    -   connect (required)
    -   main (optional, 'main' should handle files one by one)
    -   disconnect  (required)
    arguments: dict 'channel' which has all channel attributes
    more parameters/data for communicationscript:   hard code this in communicationscript; or use bots.ini
    Different ways of working:
    1. for incoming files (bots receives the files):
        1.1 connect puts all files in a directory, there is no 'main' function. bots can remove the files (if you use the 'remove' switch of the channel).
        1.2 connect only builds the connection, 'main' is a generator that passes the messages one by one (using 'yield'). bots can remove the files (if you use the 'remove' switch of the channel).
    2. for outgoing files (bots sends the files):
        2.1 if there is a 'main' function: the 'main' function is called by bots after writing each file. bots can remove the files (if you use the 'remove' switch of the channel).
        2.2 no 'main' function: the processing of all the files can be done in 'disconnect'. bots can remove the files (if you use the 'remove' switch of the channel).
    '''
    def connect(self):
        if self.userscript is None or not botslib.tryrunscript(self.userscript,self.scriptname,'connect',channeldict=self.channeldict):
            raise botslib.BotsImportError(_(u'Channel "%(idchannel)s" is type "communicationscript", but no communicationscript exists.') ,
                                {'idchannel':self.channeldict})


    @botslib.log_session
    def incommunicate(self):
        startdatetime = datetime.datetime.now()
        if hasattr(self.userscript,'main'): #process files one by one; communicationscript has to be a generator
            remove_ta = False
            for fromfilename in botslib.runscriptyield(self.userscript,self.scriptname,'main',channeldict=self.channeldict):
                try:
                    ta_from = botslib.NewTransaction(filename = fromfilename,
                                                    status = EXTERNIN,
                                                    fromchannel = self.channeldict['idchannel'],
                                                    idroute = self.idroute)
                    ta_to = ta_from.copyta(status = FILEIN)
                    remove_ta = True
                    #open fromfile
                    fromfile = open(fromfilename, 'rb')
                    filesize = os.fstat(fromfile.fileno()).st_size
                    #open tofile
                    tofilename = unicode(ta_to.idta)
                    tofile = botslib.opendata(tofilename, 'wb')
                    #copy
                    shutil.copyfileobj(fromfile,tofile,1048576)
                    fromfile.close()
                    tofile.close()
                except:
                    txt = botslib.txtexc()
                    botslib.ErrorProcess(functionname='communicationscript-incommunicate',errortext=txt,channeldict=self.channeldict)
                    if remove_ta:
                        try:
                            ta_from.delete()
                            ta_to.delete()
                        except:
                            pass
                else:
                    ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                    ta_from.update(statust=DONE)
                    if self.channeldict['remove']:
                        os.remove(fromfilename)
                finally:
                    remove_ta = False
                    if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                        break
        else:   #all files have been set ready by external communicationscript using 'connect'.
            frompath = botslib.join(self.channeldict['path'], self.channeldict['filename'])
            filelist = [filename for filename in glob.iglob(frompath) if os.path.isfile(filename)]
            filelist.sort()
            remove_ta = False
            for fromfilename in filelist:
                try:
                    ta_from = botslib.NewTransaction(filename = fromfilename,
                                                    status = EXTERNIN,
                                                    fromchannel = self.channeldict['idchannel'],
                                                    idroute = self.idroute)
                    ta_to = ta_from.copyta(status = FILEIN)
                    remove_ta = True
                    fromfile = open(fromfilename, 'rb')
                    tofilename = unicode(ta_to.idta)
                    tofile = botslib.opendata(tofilename, 'wb')
                    content = fromfile.read()
                    filesize = len(content)
                    tofile.write(content)
                    fromfile.close()
                    tofile.close()
                except:
                    txt = botslib.txtexc()
                    botslib.ErrorProcess(functionname='communicationscript-incommunicate',errortext=txt,channeldict=self.channeldict)
                    if remove_ta:
                        try:
                            ta_from.delete()
                            ta_to.delete()
                        except:
                            pass
                else:
                    ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                    ta_from.update(statust=DONE)
                    if self.channeldict['remove']:
                        os.remove(fromfilename)
                finally:
                    remove_ta = False
                    if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                        break


    @botslib.log_session
    def outcommunicate(self):
        #check if output dir exists, else create it.
        outputdir = botslib.join(self.channeldict['path'])
        botslib.dirshouldbethere(outputdir)
        #get right filename_mask & determine if fixed name (append) or files with unique names
        filename_mask = self.channeldict['filename'] if self.channeldict['filename'] else '*'
        if '{overwrite}' in filename_mask:
            filename_mask = filename_mask.replace('{overwrite}','')
            mode = 'wb'
        else:
            mode = 'ab'
        #select the db-ta's for this channel
        for row in botslib.query(u'''SELECT idta,filename,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                    AND status=%(status)s
                                    AND statust=%(statust)s
                                    AND tochannel=%(tochannel)s ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:    #for each db-ta:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to =   ta_from.copyta(status=EXTERNOUT)
                tofilename = self.filename_formatter(filename_mask,ta_from)
                #open tofile
                tofilename = botslib.join(outputdir,tofilename)
                tofile = open(tofilename, mode)
                #open fromfile
                fromfile = botslib.opendata(row['filename'], 'rb')
                #copy
                shutil.copyfileobj(fromfile,tofile,1048576)
                fromfile.close()
                tofile.close()
                #one file is written; call external
                if botslib.tryrunscript(self.userscript,self.scriptname,'main',channeldict=self.channeldict,filename=tofilename,ta=ta_from):
                    if self.channeldict['remove']:
                        os.remove(tofilename)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename=tofilename,numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        botslib.tryrunscript(self.userscript,self.scriptname,'disconnect',channeldict=self.channeldict)
        if self.channeldict['remove'] and not hasattr(self.userscript,'main'):  #if bots should remove the files, and all files are passed at once, delete these files.
            outputdir = botslib.join(self.channeldict['path'], self.channeldict['filename'])
            for filename in glob.iglob(outputdir):
                if os.path.isfile(filename):
                    try:
                        os.remove(filename)
                    except:
                        pass

class trash(_comsession):
    @botslib.log_session
    def outcommunicate(self):
        ''' does output of files to 'nothing' (trash it).
        '''
        #select the db-ta's for this channel
        for row in botslib.query(u'''SELECT idta,filename,numberofresends
                                       FROM ta
                                      WHERE idta>%(rootidta)s
                                        AND status=%(status)s
                                        AND statust=%(statust)s
                                        AND tochannel=%(tochannel)s
                                        ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:    #for each db-ta:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to =   ta_from.copyta(status=EXTERNOUT)
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename='',numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

class http(_comsession):
    scheme = 'http'     #scheme used for building url
    headers = None      #used if specified; eg {'content-type': 'application/json'}, # A dictionary with header params
    params = None      #used if specified; eg {'key1':'value1','key2':'value2'} -> http://server.com/path?key2=value2&key1=value1
    verify = False       #True, False or name of CA-file

    def connect(self):
        try:
            self.requests = botslib.botsbaseimport('requests')
        except ImportError:
            raise ImportError(_(u'Dependency failure: communicationtype "http(s)" requires python library "requests".'))
        if self.channeldict['username'] and self.channeldict['secret']:
            self.auth = (self.channeldict['username'], self.channeldict['secret'])
        else:
            self.auth = None
        self.cert = None
        self.url = botslib.Uri(scheme=self.scheme,hostname=self.channeldict['host'],port=self.channeldict['port'],path=self.channeldict['path'])

    @botslib.log_session
    def incommunicate(self):
        startdatetime = datetime.datetime.now()
        remove_ta = False
        while True:     #loop until no content is received or max communication time is expired
            try:
                #fetch via requests library
                outResponse = self.requests.get(self.url.uri(),
                                                auth=self.auth,
                                                cert=self.cert,
                                                params=self.params,
                                                headers=self.headers,
                                                verify=self.verify)
                if outResponse.status_code != self.requests.codes.ok: #communication not OK: exception
                    raise botslib.CommunicationError(_(u'%(scheme)s receive error, response code: "%(status_code)s".'),{'scheme':self.scheme,'status_code':outResponse.status_code})
                if not outResponse.content: #communication OK, but nothing received: break
                    break
                ta_from = botslib.NewTransaction(filename=self.url.uri(),
                                                    status=EXTERNIN,
                                                    fromchannel=self.channeldict['idchannel'],
                                                    idroute=self.idroute)
                ta_to =   ta_from.copyta(status=FILEIN)
                remove_ta = True
                tofilename = unicode(ta_to.idta)
                tofile = botslib.opendata(tofilename, 'wb')
                tofile.write(outResponse.content)
                tofile.close()
                filesize = len(outResponse.content)
            except:
                txt = botslib.txtexc()
                botslib.ErrorProcess(functionname='http-incommunicate',errortext=txt,channeldict=self.channeldict)
                if remove_ta:
                    try:
                        ta_from.delete()
                        ta_to.delete()
                    except:
                        pass
                break
            else:
                ta_to.update(filename=tofilename,statust=OK,filesize=filesize)
                ta_from.update(statust=DONE)
            finally:
                remove_ta = False
                if (datetime.datetime.now()-startdatetime).seconds >= self.maxsecondsperchannel:
                    break


    @botslib.log_session
    def outcommunicate(self):
        '''not used now:
            if send as 'body':
                outResponse = requests.post(url, ..., data = filedata)
            elif send as 'multipart':
                outResponse = requests.post(url, ..., files={'file': filedata})
        '''
        for row in botslib.query('''SELECT idta,filename,numberofresends
                                    FROM ta
                                    WHERE idta>%(rootidta)s
                                      AND status=%(status)s
                                      AND statust=%(statust)s
                                      AND tochannel=%(tochannel)s
                                        ''',
                                    {'tochannel':self.channeldict['idchannel'],'rootidta':self.rootidta,
                                    'status':FILEOUT,'statust':OK}):
            try:
                ta_from = botslib.OldTransaction(row['idta'])
                ta_to = ta_from.copyta(status=EXTERNOUT)
                fromfile = botslib.opendata(row['filename'], 'rb')
                content = fromfile.read()
                fromfile.close()
                #communicate via requests library
                outResponse = self.requests.post(self.url.uri(),
                                                auth=self.auth,
                                                cert=self.cert,
                                                params=self.params,
                                                headers=self.headers,
                                                data=content,
                                                verify=self.verify)
                if outResponse.status_code != self.requests.codes.ok:
                    raise botslib.CommunicationError(_(u'%(scheme)s send error, response code: "%(status_code)s".'),{'scheme':self.scheme,'status_code':outResponse.status_code})
            except:
                txt = botslib.txtexc()
                ta_to.update(statust=ERROR,errortext=txt,filename=self.url.uri(filename=row['filename']),numberofresends=row['numberofresends']+1)
            else:
                ta_to.update(statust=DONE,filename=self.url.uri(filename=row['filename']),numberofresends=row['numberofresends']+1)
            finally:
                ta_from.update(statust=DONE)

    def disconnect(self):
        pass


class https(http):
    ''' testing results:
        - http server has self-signed certificate; caCert=None, verify='/pathtocert/ca.pem': OK, certificate is verified.
        - http server has self-signed certificate; caCert points to same certificate, verify = None: OK, certificate is verified.
        - http server has self-signed certificate; verify = True: OK, but certificate is not verified.
    '''
    scheme = 'https'   #scheme used for building url
    verify = True     #verify host certificate: True, False or path to CA bundle eg '/pathtocert/ca.pem'
    caCert = None     #None or path to CA bundle eg '/pathtocert/ca.pem', Specify if https server has an unrecognized CA. Looks like verify needs to be None in order to work(??)

    def connect(self):
        #option to set environement variable for requests library; use if https server has an unrecognized CA
        super(https,self).connect()
        if self.caCert:
            os.environ["REQUESTS_CA_BUNDLE"] = self.caCert
        if self.channeldict['certfile'] and self.channeldict['keyfile']:
            self.cert = (self.channeldict['certfile'], self.channeldict['keyfile'])
        else:
            self.cert = None



import botsglobal
import models

my_context = {}     #save vars initialised at startup

def set_context(request):
    ''' set variables in the context of templates.
    '''
    global my_context
    if not my_context:
        #most context vars are from bots.ini or database. initialise these at startup
        my_context['bots_environment_text'] = botsglobal.ini.get('webserver','environment_text',' ')
        my_context['bots_environment_text_color'] = botsglobal.ini.get('webserver','environment_text_color','#000000')
        my_context['botslogo'] = botsglobal.ini.get('webserver','botslogo',"bots/botslogo.html")
        my_context['bots_touchscreen'] = botsglobal.ini.getboolean('webserver','bots_touchscreen',False)
        my_context['bots_mindate'] = 0 - botsglobal.ini.getint('settings','maxdays',30)
        my_context['menu_automaticretrycommunication'] = botsglobal.ini.getboolean('webserver','menu_automaticretrycommunication',False)
        #in bots.ini it is possible to add custom menu's
        if botsglobal.ini.has_section('custommenus'):
            my_context['custom_menuname'] = botsglobal.ini.get('custommenus','menuname','Custom')
            my_context['custom_menus'] = [(key.title(),value) for key,value in botsglobal.ini.items('custommenus') if key != 'menuname']

    #in bots.ini can be indicated that all routes (in config->routes, if route is activated) can be run individually via menu
    if botsglobal.ini.get('webserver','menu_all_routes','') == 'notindefaultrun':
        my_context['menu_all_routes'] = list(models.routes.objects.values_list('idroute', flat=True).filter(active=True).filter(notindefaultrun=True).order_by('idroute').distinct())
    elif botsglobal.ini.getboolean('webserver','menu_all_routes',False):
        my_context['menu_all_routes'] = list(models.routes.objects.values_list('idroute', flat=True).filter(active=True).order_by('idroute').distinct())

    #bots_http_path is used in name of browser-window; this is derived from url/path
    bots_http_path = request.get_full_path()
    if bots_http_path.startswith('/admin/bots/'):
        bots_http_path = bots_http_path[12:]
    else:
        bots_http_path = bots_http_path[1:]
    if bots_http_path:
        if bots_http_path[-1] == '/':
            bots_http_path = bots_http_path[:-1]
    else:
        bots_http_path = 'home'
    my_context['bots_http_path'] = bots_http_path
    my_context['bots_minDate'] = 0 - botsglobal.ini.getint('settings','maxdays',30)

    
    #***variables are set now for template use, eg {{ bots_environment_text }}
    return my_context



#!/usr/bin/env python
import sys
import os
import botsinit
import botsglobal


def start():
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #***command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    A utility to generate the index file of a plugin; this can be seen as a database dump of the configuration.
    This is eg useful for version control.
    Usage:
        %(name)s  -c<directory>
    Options:
        -c<directory>   directory for configuration files (default: config).
        
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        else:
            print usage
            sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    import pluglib              #import here, import at start of file gives error; first initialize.
    usersys = botsglobal.ini.get('directories','usersysabs')
    index_filename = os.path.join(usersys,'index.py')
    dummy_for_cleaned_data = {'databaseconfiguration':True,'umlists':botsglobal.ini.getboolean('settings','codelists_in_plugin',True),'databasetransactions':False}
    pluglib.make_index(dummy_for_cleaned_data,index_filename)
    

if __name__ == '__main__':
    start()



#!/usr/bin/env python
import sys
import os
import xmlrpclib
from SimpleXMLRPCServer import SimpleXMLRPCServer
import time
import datetime
import subprocess
import threading
import botsinit
import botslib
import botsglobal


#-------------------------------------------------------------------------------
PRIORITY = 0
JOBNUMBER = 1
TASK = 2
class Jobqueue(object):
    ''' handles the jobqueue.
        methodes can be called over xmlrpc (except the methods starting with '_')
    '''
    def __init__(self,logger):
        self.jobqueue = []       # list of jobs. in jobqueue are jobs are: (priority,jobnumber,task)
        self.jobcounter = 0      # to assign unique sequential job-number
        self.logger = logger
                
    def addjob(self,task,priority):
        #canonize task (to better find duplicates)??. Is dangerous, as non-bots-tasks might be started....
        #first check if job already in queue
        for job in self.jobqueue:
            if job[TASK] == task:
                if job[PRIORITY] != priority:   #change priority. is this useful?
                    job[PRIORITY] = priority
                    self.logger.info(u'Duplicate job, changed priority to %(priority)s: %(task)s',{'priority':priority,'task':task})
                    self._sort()
                    return 0        #zero or other code??
                else:
                    self.logger.info(u'Duplicate job not added: %(task)s',{'task':task})
                    return 4
        #add the job
        self.jobcounter += 1
        self.jobqueue.append([priority, self.jobcounter,task])
        self.logger.info(u'Added job %(job)s, priority %(priority)s: %(task)s',{'job':self.jobcounter,'priority':priority,'task':task})
        self._sort()
        return 0

    def clearjobq(self):
        self.jobqueue = []
        self.logger.info(u'Job queue cleared.')
        return 0

    def getjob(self):
        if len(self.jobqueue):
            return self.jobqueue.pop()
        return 0

    def _sort(self):
        self.jobqueue.sort(reverse=True)
        self.logger.debug(u'Job queue changed. New queue: %(queue)s',{'queue':''.join(['\n    ' + repr(job) for job in self.jobqueue])})
        #~ print self.jobqueue

#-------------------------------------------------------------------------------
def maxruntimeerror(logger,maxruntime,jobnumber,task_to_run):
    logger.error(u'Job %(job)s exceeded maxruntime of %(maxruntime)s minutes',{'job':jobnumber,'maxruntime':maxruntime})
    botslib.sendbotserrorreport(u'[Bots Job Queue] - Job exceeded maximum runtime',
                                u'Job %(job)s exceeded maxruntime of %(maxruntime)s minutes:\n %(task)s' % {'job':jobnumber,'maxruntime':maxruntime,'task':task_to_run})

#-------------------------------------------------------------------------------
def launcher(logger,port,lauchfrequency,maxruntime):
    xmlrpcclient = xmlrpclib.ServerProxy(u'http://localhost:' + unicode(port))
    maxseconds = maxruntime*60
    time.sleep(3)   #to allow jobqserver to start
    while True:
        time.sleep(lauchfrequency)
        job = xmlrpcclient.getjob()

        if job:       #0 means nothing to launch
            jobnumber = job[1]
            task_to_run = job[2]
            # Start a timer thread for maxruntime error
            timer_thread = threading.Timer(maxseconds,maxruntimeerror,args=(logger,maxruntime,jobnumber,task_to_run))
            timer_thread.start()
            try:
                starttime = datetime.datetime.now()
                logger.info(u'Starting job %(job)s',{'job':jobnumber})
                result = subprocess.call(task_to_run,stdin=open(os.devnull,'r'),stdout=open(os.devnull,'w'),stderr=open(os.devnull,'w'))
                time_taken = datetime.timedelta(seconds=(datetime.datetime.now() - starttime).seconds)
                logger.info(u'Finished job %(job)s, elapsed time %(time_taken)s, result %(result)s',{'job':jobnumber,'time_taken':time_taken,'result':result})
            except Exception as msg:
                logger.error(u'Error starting job %(job)s: %(msg)s',{'job':jobnumber,'msg':msg})
                botslib.sendbotserrorreport(u'[Bots Job Queue] - Error starting job',
                                            u'Error starting job %(job)s:\n %(task)s\n\n %(msg)s' % {'job':jobnumber,'task':task_to_run,'msg':msg})
            timer_thread.cancel()


def start():
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #***command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    Server program that ensures only a single bots-engine runs at any time, and no engine run requests are 
    lost/discarded. Each request goes to a queue and is run in sequence when the previous run completes. 
    Use of the job queue is optional and must be configured in bots.ini (jobqueue section, enabled = True).
    Usage:
        %(name)s  -c<directory>
    Options:
        -c<directory>   directory for configuration files (default: config).
        
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        else:
            print usage
            sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    if not botsglobal.ini.getboolean('jobqueue','enabled',False):
        print 'Error: bots jobqueue cannot start; not enabled in %s/bots.ini' % configdir
        sys.exit(1)
    process_name = 'jobqueue'
    logger = botsinit.initserverlogging(process_name)
    logger.log(25,u'Bots %(process_name)s started.',{'process_name':process_name})
    logger.log(25,u'Bots %(process_name)s configdir: "%(configdir)s".',{'process_name':process_name,'configdir':botsglobal.ini.get('directories','config')})
    port = botsglobal.ini.getint('jobqueue','port',28082)
    logger.log(25,u'Bots %(process_name)s listens for xmlrpc at port: "%(port)s".',{'process_name':process_name,'port':port})

    #start launcher thread
    lauchfrequency = botsglobal.ini.getint('jobqueue','lauchfrequency',5)
    maxruntime = botsglobal.ini.getint('settings','maxruntime',60)
    launcher_thread = threading.Thread(name='launcher', target=launcher, args=(logger,port,lauchfrequency,maxruntime))
    launcher_thread.daemon = True
    launcher_thread.start()
    logger.info(u'Jobqueue launcher started.')

    #the main thread is the xmlrpc server: all adding, getting etc for jobqueue is done via xmlrpc.
    logger.info(u'Jobqueue server started.')
    server = SimpleXMLRPCServer(('localhost', port),logRequests=False)        
    server.register_instance(Jobqueue(logger))
    try:
        server.serve_forever()
    except KeyboardInterrupt:
        pass
    
    sys.exit(0)


if __name__ == '__main__':
    start()



#!/usr/bin/env python
''' This source file starts up bots monitor when using apache2 as webserver.

WSGI SCRIPT EXAMPLE (outside bots directory, imports and starts this script apachewebserver.py):
import sys
import django.core.handlers.wsgi
import mod_wsgi

#Set PYTHONPATH. This is not needed if bots is already on PYTHONPATH
#sys.path.append('/usr/local/lib/python2.7/dist-packages')
from bots import apachewebserver

config = mod_wsgi.process_group
apachewebserver.start(config)
application = django.core.handlers.wsgi.WSGIHandler()
'''
'''
APACHE CONFIG FILE EXAMPLE:
WSGIScriptAlias /    <wsgi script>
Alias /media    <media directory>

Listen 8080
NameVirtualHost *:8080
<VirtualHost *:8080>
WSGIDaemonProcess config user=xxxxx
WSGIProcessGroup config
</VirtualHost>
'''
import botsglobal
import botsinit


def start(configdir):
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    process_name = 'apache_webserver_' + configdir
    botsglobal.logger = botsinit.initserverlogging(process_name)    #initialise file-logging for web-server. This logging only contains the logging from bots-webserver.





try:
    from pysqlite2 import dbapi2 as sqlite  #prefer external modules for pylite
except ImportError:
    import sqlite3 as sqlite    #works OK for python26

import re
reformatparamstyle = re.compile(u'%\((?P<name>[^)]+)\)s')

#~ #bots engine uses:
#~ ''' SELECT *
    #~ FROM ta
    #~ WHERE idta=%(idta)s ''',
    #~ {'idta':12345})
#~ #SQLite wants:
#~ ''' SELECT *
    #~ FROM ta
    #~ WHERE idta=:idta ''',
    #~ {'idta': 12345}


sqlite.register_adapter(bool, lambda s: '1' if s else '0')
sqlite.register_converter('BOOLEAN', lambda s: s == '1')

def connect(database):
    con = sqlite.connect(database, factory=BotsConnection,detect_types=sqlite.PARSE_DECLTYPES, timeout=99.0, isolation_level='EXCLUSIVE')
    con.row_factory = sqlite.Row
    con.execute('''PRAGMA synchronous=OFF''')
    return con

class BotsConnection(sqlite.Connection):
    def cursor(self):
        return sqlite.Connection.cursor(self, factory=BotsCursor)

class BotsCursor(sqlite.Cursor):
    def execute(self,string,parameters=None):
        if parameters is None:
            sqlite.Cursor.execute(self,string)
        else:
            sqlite.Cursor.execute(self,reformatparamstyle.sub(u''':\g<name>''',string),parameters)




import os
import sys
from distutils.core import setup
from distutils.command.install import INSTALL_SCHEMES

#install data file in the same way as *.py
for scheme in INSTALL_SCHEMES.values():
    scheme['data'] = scheme['purelib']


def fullsplit(path, result=None):
    '''
    Split a pathname into components (the opposite of os.path.join) in a
    platform-neutral way.
    '''
    if result is None:
        result = []
    head, tail = os.path.split(path)
    if head == '':
        return [tail] + result
    if head == path:
        return result
    return fullsplit(head, [tail] + result)


# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir != '':
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('bots'):
    # Ignore dirnames that start with '.'
    #~ for i, dirname in enumerate(dirnames):
        #~ if dirname.startswith('.'): del dirnames[i]
    if '__init__.py' in filenames:
        packages.append('.'.join(fullsplit(dirpath)))
        if len(filenames) > 1:
            data_files.append([dirpath, [os.path.join(dirpath, f) for f in filenames if not f.endswith('.pyc') and not f.endswith('.py')]])
    elif filenames:
        data_files.append([dirpath, [os.path.join(dirpath, f) for f in filenames if not f.endswith('.pyc') and not f.endswith('.py')]])








setup(
    name = 'bots-open-source-edi-translator',
    version = '3.1.8',
    author = 'hjebbers',
    author_email = 'hjebbers@gmail.com',
    url = 'http://bots.sourceforge.net/',
    description = 'Bots open source edi translator',
    long_description = 'Bots is complete software for edi (Electronic Data Interchange): translate and communicate. All major edi data formats are supported: edifact, x12, tradacoms, xml',
    platforms = 'OS Independent (Written in an interpreted language)',
    license = 'GNU General Public License (GPL)',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Office/Business',
        'Topic :: Office/Business :: Financial',
        'Topic :: Other/Nonlisted Topic',
        'Topic :: Communications',
        'Environment :: Console',
        'Environment :: Web Environment',
    ],
    scripts = [ 'bots-webserver.py',
            'bots-engine.py',
            'bots-grammarcheck.py',
            'bots-xml2botsgrammar.py',
            'bots-updatedb.py',
            'bots-dirmonitor.py',
            'bots-jobqueueserver.py',
            'bots-plugoutindex.py',
            'bots-job2queue.py',
            ],
    install_requires=[
        'django<1.8',
        'CherryPy<4.0.0',
        'Genshi<4.0.0',
    ],
    packages = packages,
    data_files = data_files,
    )



#monitors directories for new files.
#if a new file, lauch a job to the jobqueue server (so: jobqueue-server is needed).
#directories to wachs are in config/bots.ini
#runs as a daemon/service.
#this module contains separate implementations for linux and windows
#integrate in bots?? parameters should be in channels....indicate watch or not, have  path, filename. can retrieve route. add: rec

import sys
import os
import fnmatch
import threading
import time
#bots imports
import botsinit
import botsglobal
import job2queue


if os.name == 'nt':
    try:
        import win32file, win32con
    except Exception as msg: 
        raise ImportError(u'Dependency failure: bots directory monitoring requires python library "Python Win32 Extensions" on windows.')
   
    def windows_event_handler(logger,dir_watch,cond,tasks):
        ACTIONS = { 1 : "Created  ",      #tekst for printing results
                    2 : "Deleted  ",
                    3 : "Updated  ",
                    4 : "Rename from",
                    5 : "Rename to",
                    }
        FILE_LIST_DIRECTORY = 0x0001
        hDir = win32file.CreateFile(dir_watch['path'],           #path to directory
                                    FILE_LIST_DIRECTORY,          #access (read/write) mode
                                    win32con.FILE_SHARE_READ | win32con.FILE_SHARE_WRITE | win32con.FILE_SHARE_DELETE,  #share mode: FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE
                                    None,                         #security descriptor
                                    win32con.OPEN_EXISTING,       #how to create
                                    win32con.FILE_FLAG_BACKUP_SEMANTICS,    # file attributes: FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OVERLAPPED
                                    None,
                                    )
        # detecting right events is not easy in windows :-(
        # want to detect: new file,  move, drop, rename, write/append to file
        # only FILE_NOTIFY_CHANGE_LAST_WRITE: copy yes, no move
        # for rec=True: event that subdirectory itself is updated (for file deletes in dir)
        while True:
            results = win32file.ReadDirectoryChangesW(  hDir,
                                                        8192,                   #buffer size was 1024, do not want to miss anything
                                                        dir_watch['rec'],       #recursive 
                                                        win32con.FILE_NOTIFY_CHANGE_FILE_NAME |         
                                                        #~ win32con.FILE_NOTIFY_CHANGE_DIR_NAME |
                                                        #~ win32con.FILE_NOTIFY_CHANGE_ATTRIBUTES |
                                                        #~ win32con.FILE_NOTIFY_CHANGE_SIZE |
                                                        #~ win32con.FILE_NOTIFY_CHANGE_SECURITY |
                                                        #~ win32con.FILE_NOTIFY_CHANGE_CREATION |       #unknown, does not work!
                                                        #~ win32con.FILE_NOTIFsY_CHANGE_LAST_ACCESS |   #unknown, does not work!
                                                        win32con.FILE_NOTIFY_CHANGE_LAST_WRITE,
                                                        None,
                                                        None
                                                        )
            if results:
                #for each incoming event: place route to run in a set. Main thread takes action.
                for action, filename in results:
                    logger.debug(u'Event: %(action)s %(filename)s',{'action':ACTIONS.get(action,"Unknown"),'filename':filename})
                for action, filename in results:
                    if action in [1,3,5] and fnmatch.fnmatch(filename, dir_watch['filemask']):
                        #~ if dir_watch['rec'] and os.sep in filename:
                            #~ continue
                        #~ full_filename = os.path.join (path_to_watch, file)
                        cond.acquire()
                        tasks.add(dir_watch['route'])
                        cond.notify()
                        cond.release()
                        break       #the route is triggered, do not need to trigger more often
    #end of windows-specific ##################################################################################
else:
    #linux specific ###########################################################################################
    try:
        import pyinotify
    except Exception as msg: 
        raise ImportError(u'Dependency failure: bots directory monitoring requires python library "pyinotify" on linux.')
    
    class LinuxEventHandler(pyinotify.ProcessEvent):
        ''' 
        incoming event contains:
            dir=<bool>    check? - looks like the mask does nover contains dirs.
            mask=0x80
            maskname=eg IN_MOVED_TO 
            name=<filename>
            path=<path>
            pathname=<path>/<filename> 
            wd=<int>     #the watch
        ''' 
        def my_init(self, logger,dir_watch_data,cond,tasks):
            self.dir_watch_data = dir_watch_data
            self.cond = cond
            self.tasks = tasks
            self.logger = logger
            
        def process_IN_CREATE(self, event):
            ''' these events are not needed, but otherwise auto_add does not work....'''
            pass
            
        def process_default(self,event):
            ''' for each incoming event: place route to run in a set. Main thread sends actual job.
            '''
            #~ if event.mask == pyinotify.IN_CLOSE_WRITE and event.dir and self.watch_data[event.wd][2]: 
                #~ logger.info(u'new directory!!"%s %s".',event.)
            #~ print 'event detected',event.name,event.maskname, event.wd
            for dir_watch in self.dir_watch_data:
                if event.pathname.startswith(dir_watch['path']):
                    if fnmatch.fnmatch(event.name, dir_watch['filemask']):
                        self.cond.acquire()
                        self.tasks.add(dir_watch['route'])
                        self.cond.notify()
                        self.cond.release()

    def linux_event_handler(logger,dir_watch_data, cond,tasks):
        watch_manager = pyinotify.WatchManager()
        mask = pyinotify.IN_CLOSE_WRITE | pyinotify.IN_MOVED_TO | pyinotify.IN_MODIFY | pyinotify.IN_CREATE
        for dir_watch in dir_watch_data:
            watch_manager.add_watch(path=dir_watch['path'],mask=mask,rec=dir_watch['rec'],auto_add=True,do_glob=True)
        handler = LinuxEventHandler(logger=logger,dir_watch_data=dir_watch_data,cond=cond,tasks=tasks)
        notifier = pyinotify.Notifier(watch_manager, handler)
        notifier.loop()
    #end of linux-specific ##################################################################################


def start():
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #***command line arguments**************************
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    A utility to generate the index file of a plugin; this can be seen as a database dump of the configuration.
    This is eg useful for version control.
    Usage:
        %(name)s  -c<directory>
    Options:
        -c<directory>   directory for configuration files (default: config).
        
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
        else:
            print usage
            sys.exit(0)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)     #find locating of bots, configfiles, init paths etc.
    if not botsglobal.ini.getboolean('jobqueue','enabled',False):
        print 'Error: bots jobqueue cannot start; not enabled in %s/bots.ini' % configdir
        sys.exit(1)
    process_name = 'dirmonitor'
    logger = botsinit.initserverlogging(process_name)
    logger.log(25,u'Bots %(process_name)s started.',{'process_name':process_name})
    logger.log(25,u'Bots %(process_name)s configdir: "%(configdir)s".',{'process_name':process_name,'configdir':botsglobal.ini.get('directories','config')})
    
    botsenginepath = os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])),'bots-engine.py')        #get path to bots-engine
    cond = threading.Condition()
    tasks = set()    
    dir_watch_data = []
    for section in botsglobal.ini.sections():
        if section.startswith('dirmonitor') and section[len('dirmonitor'):]:
            dir_watch_data.append({})
            dir_watch_data[-1]['path'] = botsglobal.ini.get(section,'path')
            dir_watch_data[-1]['rec'] = botsglobal.ini.getboolean(section,'recursive',False)
            dir_watch_data[-1]['filemask'] = botsglobal.ini.get(section,'filemask','*')
            dir_watch_data[-1]['route'] = botsglobal.ini.get(section,'route','')
    if not dir_watch_data:
        logger.error(u'Nothing to watch!')
        sys.exit(0)

    if os.name == 'nt':
         #for windows: start a thread per directory watcher
        for dir_watch in dir_watch_data:
            dir_watch_thread = threading.Thread(target=windows_event_handler, args=(logger,dir_watch,cond,tasks))
            dir_watch_thread.daemon = True  #do not wait for thread when exiting
            dir_watch_thread.start()
    else:
        #for linux: one watch-thread, but multiple watches. 
        dir_watch_thread = threading.Thread(target=linux_event_handler, args=(logger,dir_watch_data,cond,tasks))
        dir_watch_thread.daemon = True  #do not wait for thread when exiting
        dir_watch_thread.start()

    # this main thread get the results from the watch-thread(s).
    logger.info(u'Bots %(process_name)s started.',{'process_name':process_name})
    active_receiving = False
    timeout = 2.0
    cond.acquire()
    while True:
        #this functions as a buffer: all events go into set tasks.
        #the tasks are fired to jobqueue after TIMOUT sec.
        #this is to avoid firing to many tasks to jobqueue; events typically come in bursts.
        #is value of timeout is larger, reaction times are slower...but less tasks are fired to jobqueue.
        #in itself this is not a problem, as jobqueue will alos discard duplicate jobs.
        #2 sec seems to e a good value: reasonable quick, not to nervous. 
        cond.wait(timeout=timeout)    #get back when results, or after timeout sec
        if tasks:
            if not active_receiving:    #first request (after tasks have been  fired, or startup of dirmonitor)
                active_receiving = True
                last_time = time.time()
            else:     #active receiving events
                current_time = time.time()
                if current_time - last_time >= timeout:  #cond.wait returned probably because of a timeout
                    try:
                        for task in tasks:
                            logger.info(u'Send to queue "%(path)s %(config)s %(task)s".',{'path':botsenginepath,'config':'-c' + configdir,'task':task})
                            job2queue.send_job_to_jobqueue([sys.executable,botsenginepath,'-c' + configdir,task])
                    except Exception as msg:
                        logger.info(u'Error in running task: "%(msg)s".',{'msg':msg})
                    tasks.clear()
                    active_receiving = False
                else:                                   #cond.wait returned probably because of a timeout
                    logger.debug(u'time difference to small.')
                    last_time = current_time
    cond.release()
    sys.exit(0)


if __name__ == '__main__':
    start()



#!/usr/bin/env python
import sys
import os
import xmlrpclib
import socket
import botsinit
import botsglobal

JOBQUEUEMESSAGE2TXT = {
    0: u'OK, job is added to queue',
    1: u'Error, job not to jobqueue. Can not contact jobqueue-server',
    4: u'Duplicate job, not added.',
    }


def send_job_to_jobqueue(task_args,priority=5):
    ''' adds a new job to the bots-jobqueueserver.
        is an xmlrpc client.
        Import this function in eg views.py.
        Received return codes  from jobqueueserver:
        0 = OK, job added to job queue.
        4 = job is a duplicate of job already in the queue
    '''
    try:
        remote_server = xmlrpclib.ServerProxy(u'http://localhost:' + unicode(botsglobal.ini.getint('jobqueue','port',28082)))
        return remote_server.addjob(task_args,priority)
    except socket.error as msg:
        print 'socket.error',msg
        return 1    #jobqueueserver server not active


def start():
    #NOTE: bots directory should always be on PYTHONPATH - otherwise it will not start.
    #***command line arguments**************************
    #if config (-c option) is before job argument, will be used as config-dir of job2queue it self and not for job
    #if config (-c option) is after job argument, use both as config-dir of job2queue and as -c option of job.
    #if config (-c option) is before and after job argument, use only the after...could change that but seems not to be useful.
    usage = '''
    This is "%(name)s" version %(version)s, part of Bots open source edi translator (http://bots.sourceforge.net).
    Places a job in the bots jobqueue. Bots jobqueue takes care of correct processing of jobs.    
    Usage:
        %(name)s  [-c<directory>] [-p<priority>] job [job-parameters]
    Options:
        -c<directory>   directory for configuration files (default: config).
        -p<priority>    priority of job, 1-9 (default: 5, highest priority is 1).
    Example of usage:
        %(name)s bots-engine.py
        %(name)s python2.7 /usr/local/bin/bots-engine.py
        %(name)s -p1 python2.7 /usr/local/bin/bots-engine.py -cconfig2 myroute
        
    '''%{'name':os.path.basename(sys.argv[0]),'version':botsglobal.version}
    configdir = 'config'    #default value
    priority = 5            #default value
    task_args = []
    for arg in sys.argv[1:]:
        if arg.startswith('-c'):
            configdir = arg[2:]
            if not configdir:
                print 'Error: configuration directory indicated, but no directory name.'
                sys.exit(1)
            if task_args:
                task_args.append(arg)
        elif arg.startswith('-p'):
            try:
                priority =  int(arg[2:])
            except:
                print 'Error: priority should be numeric (1=highest, 9=lowest).'
                sys.exit(1)
        elif arg in ["?", "/?",'-h', '--help']:
            print usage
            sys.exit(0)
        else:
            task_args.append(arg)
    #***end handling command line arguments**************************
    botsinit.generalinit(configdir)         #needed to read config
    if not botsglobal.ini.getboolean('jobqueue','enabled',False):
        print 'Error: bots jobqueue cannot start; not enabled in %s/bots.ini' % configdir
        sys.exit(1)
        
    terug = send_job_to_jobqueue(task_args,priority)
    print JOBQUEUEMESSAGE2TXT[terug]
    sys.exit(terug)


if __name__ == '__main__':
    start()



try:
    import cdecimal as decimal
except ImportError:
    import decimal
from django.utils.translation import ugettext as _
import botslib
import botsglobal
from botsconfig import *


class Node(object):
    ''' Node class for building trees in inmessage and outmessage
    '''
    #slots: python optimalisation to preserve memory. Disadv.: no dynamic attr in this class
    #in tests: for normal translations less memory and faster; no effect fo one-on-one translations.
    __slots__ = ('record','children','_queries','linpos_info','structure')
    def __init__(self,record=None,linpos_info=None):
        if record and 'BOTSIDnr' not in record:
            record['BOTSIDnr'] = u'1'
        self.record = record    #record is a dict with fields
        self.children = []
        self.linpos_info = linpos_info
        self._queries = None
        self.structure = None

    def linpos(self):
        if self.linpos_info:
            return ' line %(lin)s pos %(pos)s'%{'lin':self.linpos_info[0],'pos':self.linpos_info[1]}
        else:
            return ''

    def append(self,childnode):
        '''append child to node'''
        self.children.append(childnode)

    #********************************************************
    #*** queries ********************************************
    #********************************************************
    def _getquerie(self):
        ''' getter for queries: get queries of a node '''
        if self._queries:
            return self._queries
        else:
            return {}

    def _updatequerie(self,updatequeries):
        ''' setter for queries: set/update queries of a node with dict queries.
        '''
        if updatequeries:
            if self._queries is None:
                self._queries = updatequeries.copy()
            else:
                botslib.updateunlessset(self._queries,updatequeries)

    queries = property(_getquerie,_updatequerie)

    def processqueries(self,queries,maxlevel):
        ''' copies values for queries 'down the tree' untill right level.
            So when edi file is split up in messages,
            querie-info from higher levels is copied to message.'''
        self.queries = queries
        if self.record and not maxlevel:
            return
        for childnode in self.children:
            childnode.processqueries(self.queries,maxlevel-1)

    def displayqueries(self,level=0):
        '''for debugging
            usage: in mappingscript:     inn.root.displayqueries()
        '''
        if level == 0:
            print 'displaying queries for nodes in tree'
        print '    '*level,'node:',
        if self.record:
            print self.record['BOTSID'],
        else:
            print 'None',
        print '',
        print self.queries
        for childnode in self.children:
            childnode.displayqueries(level+1)

    #********************************************************
    #*** handling of QUERIES and SUBTRANSLATION *************
    #********************************************************
    def enhancedget(self,mpaths):
        ''' to get QUERIES or SUBTRANSLATION;
            mpath can be
            - dict:     do get(mpath); can not be a mpath with multiple
            - tuple:    do get(mpath); can be multiple dicts in mapth
            - list:     for each listmembr do a get(); append the results
            - function: function should return a value (or None, if not found)
        '''
        if isinstance(mpaths,dict):
            return self.get(mpaths)
        elif isinstance(mpaths,tuple):
            return self.get(*mpaths)
        elif isinstance(mpaths,list):
            collect = u''
            for mpath in mpaths:
                if isinstance(mpath,dict):
                    found = self.get(mpath)
                elif isinstance(mpath,tuple):
                    found = self.get(*mpath)
                else:
                    raise botslib.MappingFormatError(_(u'Member in list %(mpath)s must be dict or tuple (in enhancedget).'),{'mpath':mpaths})
                if found:
                    collect += found
            return collect
        elif callable(mpaths):
            return mpaths(thisnode=self)
        else:
            raise botslib.MappingFormatError(_(u'Must be dict, list or tuple: enhancedget(%(mpath)s)'),{'mpath':mpaths})

    def get_queries_from_edi(self,record_definition):
        ''' extract information from edifile using QUERIES in grammar.structure; information will be placed in ta_info and in db-ta
        '''
        tmpdict = {}
        for key,value in record_definition[QUERIES].iteritems():
            found = self.enhancedget(value)   #search in last added node
            if found is not None:
                found = found.strip()         #needed to get correct ISA-fields
                if found:
                    tmpdict[key] = found
        self.queries = tmpdict

    #********************************************************
    #*** manipulate node.tree: get, put, change, delete etc *
    #********************************************************
    def getrecord(self,*mpaths):
        ''' get whole node record
        '''
        self._mpath_sanity_check(mpaths)
        for part in mpaths:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        terug =  self._getrecordcore(mpaths)
        botsglobal.logmap.debug(u'"%(terug)s" for getrecord%(mpaths)s',{'terug':unicode(terug),'mpaths':unicode(mpaths)})
        return terug

    def _getrecordcore(self,mpaths):
        for key,value in mpaths[0].iteritems():       #check all key, value for first part of where;
            if key not in self.record or value != self.record[key]:
                return None    #no match:
        else:   #all key,value are matched.
            if len(mpaths) == 1:    #mpath is exhausted; so we are there!!! #replace values with values in 'change'; delete if None
                return self.record
            else:           #go recursive
                for childnode in self.children:
                    terug = childnode._getrecordcore(mpaths[1:])
                    if terug:
                        return terug
                else:   #no child has given a valid return
                    return None

    def change(self,where,change):
        ''' where: node to find.
            change is applied to found node
            uses first matching node for 'where'.
        '''
        self._mpath_sanity_check(where)
        for part in where:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        #sanity check 'change' parameter
        if not isinstance(change,dict):
            raise botslib.MappingFormatError(_(u'Parameter "change" must be dict: change(where=%(where)s,change=%(change)s)'),
                                                {'where':where,'change':change})
        change.pop('BOTSID','nep')  #remove 'BOTSID' from change. BOTSID can not be changed
        change.pop('BOTSIDnr','nep')  #remove 'BOTSIDnr' from change. BOTSIDnr can not be changed
        for key,value in change.iteritems():
            if not isinstance(key,basestring):
                raise botslib.MappingFormatError(_(u'Keys in "change" must be strings: change(where=%(where)s,change=%(change)s)'),
                                                    {'where':where,'change':change})
            if not isinstance(value,basestring) and value is not None:     #if None, item is deleted
                raise botslib.MappingFormatError(_(u'Values in "change" must be strings or "None": change(where=%(where)s,change=%(change)s)'),
                                                    {'where':where,'change':change})
        terug =  self._changecore(where,change)
        botsglobal.logmap.debug(u'"%(terug)s" for change(where=%(where)s,change=%(change)s)',{'terug':terug,'where':unicode(where),'change':unicode(change)})
        return terug

    def _changecore(self,where,change):
        for key,value in where[0].iteritems():       #check all key, value for first part of where;
            if key not in self.record or value != self.record[key]:
                return False    #no match:
        else:   #all key,value are matched.
            if len(where) == 1:    #mpath is exhausted; so we are there!!! #replace values with values in 'change'; delete if None
                for key,value in change.iteritems():
                    if value is None:
                        self.record.pop(key,'nep')
                    else:
                        self.record[key] = value
                return True
            else:           #go recursive
                for childnode in self.children:
                    if childnode._changecore(where[1:],change):
                        return True
                else:   #no child has given a valid return
                    return False

    def delete(self,*mpaths):
        ''' delete the last record of mpath if found (first: find/identify record, than delete).'''
        self._mpath_sanity_check(mpaths)
        if len(mpaths) == 1:
            raise botslib.MappingFormatError(_(u'Only one dict: not allowed. Use different solution: delete(%(mpath)s)'),{'mpath':mpaths})
        for part in mpaths:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        terug =  bool(self._deletecore(mpaths))
        botsglobal.logmap.debug(u'"%(terug)s" for delete%(mpaths)s',{'terug':terug,'mpaths':unicode(mpaths)})
        return terug  #return False if not removed, return True if removed

    def _deletecore(self,mpaths):
        for key,value in mpaths[0].iteritems():          #check all items in first part of mpaths
            if key not in self.record or value != self.record[key]:
                return 0
        else:   #all items are matched
            if len(mpaths) == 1:    #mpath is exhausted; so we are there!!!
                return 2  #indicates node should be removed
            else:
                for i, childnode in enumerate(self.children):
                    terug =  childnode._deletecore(mpaths[1:]) #search recursive for rest of mpaths
                    if terug == 2:  #indicates node should be removed
                        del self.children[i]    #remove node
                        return 1    #this indicates: deleted successfull, do not remove anymore (no removal of parents)
                    if terug:
                        return terug
                else:   #no child has given a valid return
                    return 0

    def get(self,*mpaths):
        ''' get value of a field in a record from a edi-message
            mpath is xpath-alike query to identify the record/field
            function returns 1 value; return None if nothing found.
            if more than one value can be found: first one is returned
            starts searching in current node, then deeper
        '''
        if Node.checklevel:
            self._mpath_sanity_check(mpaths[:-1])
            #sanity check of last part of mpaths: None only allowed in last section of Mpath; check last part
            if not isinstance(mpaths[-1],dict):
                raise botslib.MappingFormatError(_(u'Must be dicts in tuple: get(%(mpath)s)'),{'mpath':mpaths})
            if 'BOTSID' not in mpaths[-1]:
                raise botslib.MappingFormatError(_(u'Last section without "BOTSID": get(%(mpath)s)'),{'mpath':mpaths})
            count = 0
            for key,value in mpaths[-1].iteritems():
                if not isinstance(key,basestring):
                    raise botslib.MappingFormatError(_(u'Keys must be strings in last section: get(%(mpath)s)'),{'mpath':mpaths})
                if value is None:
                    count += 1
                elif not isinstance(value,basestring):
                    raise botslib.MappingFormatError(_(u'Values must be strings (or none) in last section: get(%(mpath)s)'),{'mpath':mpaths})
            if count > 1:
                raise botslib.MappingFormatError(_(u'Max one "None" in last section: get(%(mpath)s)'),{'mpath':mpaths})
        for part in mpaths:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        if Node.checklevel == 2:
            self._mpath_grammar_check(mpaths)
        terug =  self._getcore(mpaths)
        botsglobal.logmap.debug(u'"%(terug)s" for get%(mpaths)s',{'terug':terug,'mpaths':unicode(mpaths)})
        return terug

    def _getcore(self,mpaths):
        if len(mpaths) != 1:    #node is not end-node
            for key,value in mpaths[0].iteritems():          #check all items in mpath;
                if key not in self.record or value != self.record[key]:  #does not match/is not right node
                    return None
            else:   #all items in mpath are matched and OK; recursuve search
                for childnode in self.children:
                    terug =  childnode._getcore(mpaths[1:]) #recursive search for rest of mpaths
                    if terug is not None:
                        return terug
                else:   #nothing found in children
                    return None
        else:   #node is end-node
            terug = 1 #default return value: if there is no 'None' in the mpath, but everything is matched, 1 is returned (like True)
            for key,value in mpaths[0].iteritems():          #check all items in mpath;
                if key not in self.record:  #does not match/is not right node
                    return None
                elif value is None:   #item has None-value; remember this value because this might be the right node
                    terug = self.record[key][:]    #copy to avoid memory problems
                elif value != self.record[key]:   #compare values
                    return None     #does not match/is not right node
            else:   #all keys/values in this mpathr are matched and OK.
                return terug        #either the remembered value is returned or 1 (as a boolean, indicated 'found)


    def getcount(self):
        '''count the number of nodes/records under the node/in whole tree'''
        count = 0
        if self.record:
            count += 1   #count itself
        for childnode in self.children:
            count += childnode.getcount()
        return count

    def getcountoccurrences(self,*mpaths):
        ''' count number of occurences of mpath. Eg count nr of LIN's'''
        return len(list(self.getloop(*mpaths)))

    def getcountsum(self,*mpaths):
        ''' return the sum for all values found in mpath. Eg total number of ordered quantities.'''
        mpath_for_found_node = mpaths[-1].copy()
        for key,value in mpaths[-1].items():
            if value is None:
                del mpaths[-1][key]
        count = decimal.Decimal(0)
        for i in self.getloop(*mpaths):
            count += i.getdecimal(mpath_for_found_node)
        return unicode(count)

    def getloop(self,*mpaths):
        ''' generator. Returns one by one the nodes as indicated in mpath
        '''
        if Node.checklevel:
            self._mpath_sanity_check(mpaths)
        for part in mpaths:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        if Node.checklevel == 2:
            self._mpath_grammar_check(mpaths)
        for terug in self._getloopcore(mpaths):
            botsglobal.logmap.debug(u'getloop %(mpaths)s returns "%(record)s".',{'mpaths':mpaths,'record':terug.record})
            yield terug

    def _getloopcore(self,mpaths):
        ''' recursive part of getloop()
        '''
        for key,value in mpaths[0].iteritems():
            if key not in self.record or value != self.record[key]:
                return
        else:   #all items are checked and OK.
            if len(mpaths) == 1:
                yield self      #found!
            else:
                for childnode in self.children:
                    for terug in childnode._getloopcore(mpaths[1:]): #search recursive for rest of mpaths
                        yield terug

    def getnozero(self,*mpaths):
        ''' like get, but either return a numerical value (as string) or None. If value to return is equal to zero, None is returned.
            useful eg for fixed records, where num fields are initialised with zeroes.
        '''
        terug = self.get(*mpaths)
        try:
            value = float(terug)
        except (TypeError,ValueError):
            return None
        if value == 0:
            return None
        return terug

    def getdecimal(self,*mpaths):
        ''' like get, but return a decimal value. If value to return is None or not numeric, zero is returned.
            warning: this function might 'mask' errors, because when input is not-valid, it aalwyas return a decimal 0. 
            useful eg for when math calculations are needed in mapping.
        '''
        terug = self.get(*mpaths)
        if terug and terug[-1] == u'-':    #if minus-sign at the end, put it in front. This is useful for idocs, where fields are never defined as numeric.
            terug = terug[-1] + terug[:-1]
        try:
            return decimal.Decimal(terug)
        except (TypeError,ValueError):
            return decimal.Decimal('0')


    def put(self,*mpaths,**kwargs):
        #sanity check of mpaths
        if not mpaths or not isinstance(mpaths,tuple):
            raise botslib.MappingFormatError(_(u'Must be dicts in tuple: put(%(mpath)s)'),{'mpath':mpaths})
        for part in mpaths:
            if not isinstance(part,dict):
                raise botslib.MappingFormatError(_(u'Must be dicts in tuple: put(%(mpath)s)'),{'mpath':mpaths})
            if 'BOTSID' not in part:
                raise botslib.MappingFormatError(_(u'Section without "BOTSID": put(%(mpath)s)'),{'mpath':mpaths})
            for key,value in part.iteritems():
                if value is None:
                    botsglobal.logmap.debug(u'"None" in put %(mpaths)s.',{'mpaths':unicode(mpaths)})
                    return False
                if not isinstance(key,basestring):
                    raise botslib.MappingFormatError(_(u'Keys must be strings: put(%(mpath)s)'),{'mpath':mpaths})
                if isinstance(value,list):
                    #empty is not useful, drop it (like None)
                    if not value:
                        botsglobal.logmap.debug(u'Empty list in put %(mpaths)s.',{'mpaths':unicode(mpaths)})
                        return False
                else:
                    if kwargs.get('strip',True):
                        part[key] = unicode(value).strip()  #leading and trailing spaces are stripped from the values
                    else:
                        part[key] = unicode(value)          #used for fixed ISA header of x12
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'

        if self._sameoccurence(mpaths[0]):
            self._putcore(mpaths[1:])
        else:
            raise botslib.MappingRootError(_(u'Error in root put "%(mpath)s".'),{'mpath':mpaths[0]})
        botsglobal.logmap.debug(u'"True" for put %(mpaths)s',{'mpaths':unicode(mpaths)})
        return True

    def _putcore(self,mpaths):
        if not mpaths:  #newmpath is exhausted, stop searching.
            return
        for childnode in self.children:
            if childnode.record['BOTSID'] == mpaths[0]['BOTSID'] and childnode._sameoccurence(mpaths[0]):    #checking of BOTSID is also done in sameoccurance!->performance!
                childnode._putcore(mpaths[1:])
                return
        else:   #is not present in children, so append mpath as a new node
            self.append(Node(mpaths[0]))
            self.children[-1]._putcore(mpaths[1:])

    def putloop(self,*mpaths):
        #sanity check of mpaths
        if not mpaths or not isinstance(mpaths,tuple):
            raise botslib.MappingFormatError(_(u'Must be dicts in tuple: putloop(%(mpath)s)'),{'mpath':mpaths})
        for part in mpaths:
            if not isinstance(part,dict):
                raise botslib.MappingFormatError(_(u'Must be dicts in tuple: putloop(%(mpath)s)'),{'mpath':mpaths})
            if 'BOTSID' not in part:
                raise botslib.MappingFormatError(_(u'Section without "BOTSID": putloop(%(mpath)s)'),{'mpath':mpaths})
            for key,value in part.iteritems():
                if not isinstance(key,basestring):
                    raise botslib.MappingFormatError(_(u'Keys must be strings: putloop(%(mpath)s)'),{'mpath':mpaths})
                if value is None:
                    return False
                part[key] = unicode(value).strip()
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        if self._sameoccurence(mpaths[0]):
            if len(mpaths)==1:
                return self
            return self._putloopcore(mpaths[1:])
        else:
            raise botslib.MappingRootError(_(u'Error in root putloop "%(mpath)s".'),{'mpath':mpaths[0]})

    def _putloopcore(self,mpaths):
        if len(mpaths) ==1: #end of mpath reached; always make new child-node
            self.append(Node(mpaths[0]))
            return self.children[-1]
        for childnode in self.children:  #if first part of mpaths exists already in children go recursive
            if childnode.record['BOTSID'] == mpaths[0]['BOTSID'] and childnode._sameoccurence(mpaths[0]):    #checking of BOTSID is also done in sameoccurance!->performance!
                return childnode._putloopcore(mpaths[1:])
        else:   #is not present in children, so append a child, and go recursive
            self.append(Node(mpaths[0]))
            return self.children[-1]._putloopcore(mpaths[1:])

    def _sameoccurence(self, mpath):
        ''' checks if all items that appear in both node and mpath have the same value. If so, all new items in mpath are added to node
        '''
        for key,value in self.record.iteritems():
            if key in mpath and mpath[key] != value:
                return False
        else:    #all equal keys have same values, thus both are 'equal'.
            self.record.update(mpath)   #add items to self.record that are new
            return True

    def sort(self,*mpaths):
        ''' sort nodes. eg in mappingscript:     inn.sort({'BOTSID':'UNH'},{'BOTSID':'LIN','C212.7140':None})
            This will sort the LIN segments by article number.
        '''
        if Node.checklevel:
            self._mpath_sanity_check(mpaths[:-1])
            #sanity check of last part of mpaths: None only allowed in last section of Mpath; check last part
            if not isinstance(mpaths[-1],dict):
                raise botslib.MappingFormatError(_(u'Must be dicts in tuple: get(%(mpath)s)'),{'mpath':mpaths})
            if 'BOTSID' not in mpaths[-1]:
                raise botslib.MappingFormatError(_(u'Last section without "BOTSID": get(%(mpath)s)'),{'mpath':mpaths})
            count = 0
            for key,value in mpaths[-1].iteritems():
                if not isinstance(key,basestring):
                    raise botslib.MappingFormatError(_(u'Keys must be strings in last section: get(%(mpath)s)'),{'mpath':mpaths})
                if value is None:
                    count += 1
                elif not isinstance(value,basestring):
                    raise botslib.MappingFormatError(_(u'Values must be strings (or none) in last section: get(%(mpath)s)'),{'mpath':mpaths})
            if count > 1:
                raise botslib.MappingFormatError(_(u'Max one "None" in last section: get(%(mpath)s)'),{'mpath':mpaths})
        for part in mpaths:
            if 'BOTSIDnr' not in part:
                part['BOTSIDnr'] = u'1'
        if Node.checklevel == 2:
            self._mpath_grammar_check(mpaths)
        comparekey = mpaths[1:]
        #does not work with checklevel == 2, so temporary set checklevel to 0 
        remember = Node.checklevel
        Node.checklevel = 0
        try:
            self.children.sort(key=lambda s: s.get(*comparekey))
        finally:
            Node.checklevel = remember
    #********************************************************
    #*** utility functions **********************************
    #********************************************************
    @staticmethod
    def _mpath_sanity_check(mpaths):
        ''' sanity check of mpaths. '''
        if not isinstance(mpaths,tuple):
            raise botslib.MappingFormatError(_(u'Parameter mpath must be tuple: %(mpaths)s'),{'mpaths':mpaths})
        for part in mpaths:
            if not isinstance(part,dict):
                raise botslib.MappingFormatError(_(u'Parameter mpath must be dicts in a tuple: %(mpaths)s'),{'mpaths':mpaths})
            if 'BOTSID' not in part:
                raise botslib.MappingFormatError(_(u'"BOTSID" is required in mpath: %(mpaths)s'),{'mpaths':mpaths})
            for key,value in part.iteritems():
                if not isinstance(key,basestring):
                    raise botslib.MappingFormatError(_(u'Keys must be strings in mpath: %(mpaths)s'),{'mpaths':mpaths})
                if not isinstance(value,basestring):
                    raise botslib.MappingFormatError(_(u'Values must be strings in mpath: getrecord(%(mpaths)s)'),{'mpaths':mpaths})

    def _mpath_grammar_check(self,mpaths):
        ''' check of mpaths with grammar. '''
        def _mpath_ok_with_grammar(structure,mpaths):
            ''' inner function, recursive.
                every part of mpaths should be in structure, at right level, have right fields.
            '''
            mpath = mpaths[0]
            for record_definition in structure:
                if record_definition[ID] == mpath['BOTSID'] and record_definition[BOTSIDNR] == mpath['BOTSIDnr']:
                    for key in mpath:
                        if key == 'BOTSIDnr':   #BOTSIDnr is not in grammar, so do not check
                            continue
                        for field_definition in record_definition[FIELDS]:
                            if field_definition[ISFIELD]:
                                if key == field_definition[ID]:
                                    break   #check next key
                            else:
                                if key == field_definition[ID]:         #first check compostie-key itself, for repeating composites
                                    break   #check next key
                                for grammarsubfield in field_definition[SUBFIELDS]:   #loop subfields
                                    if key == grammarsubfield[ID]:
                                        break   #check next key
                                else:
                                    continue #checking field_defintions
                                break   #check next key
                        else:   #Not found in record!
                            return False
                    #all fields in mpath are correct; go to next level of mpath..
                    if mpaths[1:]:
                        if not LEVEL in record_definition:
                            return False
                        return _mpath_ok_with_grammar(record_definition[LEVEL],mpaths[1:])
                    else:
                        return True    #no more levels, all fields found
            else:
                return False
        if not self.structure:
            return
        if not _mpath_ok_with_grammar([self.structure],mpaths):
            raise botslib.MappingFormatError(_(u'Parameter mpath is not valid according to grammar: %(mpaths)s'),{'mpaths':mpaths})

    def display(self,level=0):
        '''for debugging
            usage eg in mappingscript:     inn.root.display()
        '''
        if level == 0:
            print 'displaying all nodes in node tree:'
        print '    '*level,self.record
        for childnode in self.children:
            childnode.display(level+1)

    def stripnode(self):
        ''' removes spaces from all fields in tree.
            used in alt translations where a dict is returned by mappingscript indicating the out-tree should be used as inn.
            the out-tree has been formatted already, this is not OK for fixed formats (idoc!)
        '''
        if self.record is not None:
            for key, value in self.record.iteritems():
                self.record[key] = value.strip()
        for child in self.children:
            child.stripnode()

    def collectlines(self,print_as_row):
        ''' for new edifax routine: collect same nodes in Node.children in a list; for table-row printing.
        ''' 
        new = []        #new list
        for childnode in self.children:
            if childnode.structure[MPATH] in print_as_row:
                if not new:     #if new is still empty
                    new.append([childnode])
                elif isinstance(new[-1],list) and new[-1][0].record['BOTSID'] == childnode.record['BOTSID']:
                    new[-1].append(childnode)       #append in tabel-row-list
                else:           #other recordID, also print as table-row
                    new.append([childnode])
            else:               #just append, no change
                new.append(childnode)
                childnode.collectlines(print_as_row)   #go recursive
        self.children = new

    def copynode(self):
        ''' make a 'safe' copy of node; return the new node
        '''
        if self.record is None:
            new_node = Node(record=None)
        else:
            new_node = Node(record=dict(self.record))
        for childnode in self.children:
            new_node.append(childnode.copynode())
        return new_node
